from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_delete, post_save, pre_delete
from django.dispatch import receiver
from django.utils import timezone

from edc_constants.constants import NO
from edc_utils import formatted_datetime
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.utils import get_related_visit_model_cls

from ..appointment_status_updater import (
    AppointmentStatusUpdater,
    AppointmentStatusUpdaterError,
)
from ..constants import IN_PROGRESS_APPT, NEW_APPT
from ..creators import create_next_appointment_as_interim
from ..managers import AppointmentDeleteError
from ..model_mixins import NextAppointmentCrfModelMixin
from ..skip_appointments import SkipAppointments
from ..utils import (
    cancelled_appointment,
    get_allow_skipped_appt_using,
    get_appointment_model_name,
    missed_appointment,
    reset_visit_code_sequence_or_pass,
)
from .appointment import Appointment


@receiver(post_save, sender=Appointment, weak=False, dispatch_uid="appointment_post_save")
def appointment_post_save(sender, instance, raw, update_fields, **kwargs):
    if not raw and not update_fields:
        # use the AppointmentStatusUpdater to set all
        # other appointments to be NOT in progress
        try:
            if instance.appt_status == IN_PROGRESS_APPT:
                AppointmentStatusUpdater(
                    instance,
                    change_to_in_progress=True,
                    clear_others_in_progress=True,
                )
        except AttributeError as e:
            if "appt_status" not in str(e):
                raise
        except AppointmentStatusUpdaterError:
            pass
    # don't indent!
    # try to create_missed_visit_from_appointment
    #  if appt_status == missed
    missed_appointment(instance)
    # try to delete subject visit
    # if appt status = CANCELLED_APPT
    cancelled_appointment(instance)


@receiver(post_save, weak=False, dispatch_uid="create_appointments_on_post_save")
def create_appointments_on_post_save(sender, instance, raw, created, using, **kwargs):
    """Method `Model.create_appointments` is not typically used.

    See schedule.put_on_schedule() in edc_visit_schedule.
    """
    if not raw and not kwargs.get("update_fields"):
        try:
            instance.create_appointments()
        except AttributeError as e:
            if "create_appointments" not in str(e):
                raise


@receiver(post_save, weak=False, dispatch_uid="update_appt_status_on_related_visit_post_save")
def update_appt_status_on_related_visit_post_save(
    sender, instance, created, raw, update_fields, **kwargs
):
    if (
        not raw
        and not update_fields
        and created
        and isinstance(instance, (get_related_visit_model_cls(),))
    ):
        try:
            AppointmentStatusUpdater(
                instance.appointment,
                change_to_in_progress=True,
                clear_others_in_progress=True,
            )
        except AttributeError as e:
            if "appointment" not in str(e):
                raise
        except AppointmentStatusUpdaterError:
            pass


@receiver(
    post_delete,
    weak=False,
    dispatch_uid="update_appt_status_on_related_visit_post_delete",
)
def update_appt_status_on_related_visit_post_delete(sender, instance, using, **kwargs):
    if isinstance(instance, (get_related_visit_model_cls(),)):
        try:
            appointment = instance.appointment
        except AttributeError as e:
            if "appointment" not in str(e):
                raise
        else:
            appointment.appt_status = NEW_APPT
            appointment.save_base(update_fields=["appt_status"])


@receiver(
    pre_delete,
    sender=Appointment,
    weak=False,
    dispatch_uid="appointments_on_pre_delete",
)
def appointments_on_pre_delete(sender, instance, using, **kwargs):
    if instance.visit_code_sequence == 0:
        schedule = site_visit_schedules.get_visit_schedule(
            instance.visit_schedule_name
        ).schedules.get(instance.schedule_name)
        onschedule_datetime = schedule.onschedule_model_cls.objects.get(
            subject_identifier=instance.subject_identifier
        ).onschedule_datetime
        # get visits for this consent/consent ext
        visits = schedule.visits_for_subject(
            subject_identifier=instance.subject_identifier,
            report_datetime=onschedule_datetime,
            site_id=instance.site_id,
        )
        if instance.visit_code in [visit for visit in visits]:
            try:
                offschedule_datetime = schedule.offschedule_model_cls.objects.get(
                    subject_identifier=instance.subject_identifier
                ).offschedule_datetime
            except ObjectDoesNotExist as e:
                raise AppointmentDeleteError(
                    f"Appointment may not be deleted. "
                    f"Subject {instance.subject_identifier} is on schedule "
                    f"'{instance.visit_schedule.verbose_name}.{instance.schedule_name}' "
                    f"as of '{formatted_datetime(onschedule_datetime)}'. "
                    f"Got appointment {instance.visit_code}.{instance.visit_code_sequence} "
                    f"datetime {formatted_datetime(instance.appt_datetime)}. "
                    f"Perhaps complete off schedule model "
                    f"'{instance.schedule.offschedule_model_cls().verbose_name.title()}' "
                    f"first."
                ) from e
            else:
                if onschedule_datetime <= instance.appt_datetime <= offschedule_datetime:
                    raise AppointmentDeleteError(
                        f"Appointment may not be deleted. "
                        f"Subject {instance.subject_identifier} is on schedule "
                        f"'{instance.visit_schedule.verbose_name}.{instance.schedule_name}' "
                        f"as of '{formatted_datetime(onschedule_datetime)}' "
                        f"until '{formatted_datetime(timezone.now())}'. "
                        f"Got appointment datetime "
                        f"{formatted_datetime(instance.appt_datetime)}. "
                    )


@receiver(
    post_delete,
    sender=Appointment,
    weak=False,
    dispatch_uid="appointments_on_post_delete",
)
def appointments_on_post_delete(sender, instance, using, **kwargs):
    if (
        not kwargs.get("update_fields")
        and sender._meta.label_lower == get_appointment_model_name()
    ):
        reset_visit_code_sequence_or_pass(
            subject_identifier=instance.subject_identifier,
            visit_schedule_name=instance.visit_schedule_name,
            schedule_name=instance.schedule_name,
            visit_code=instance.visit_code,
        )


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_appointments_to_next_on_post_save",
)
def update_appointments_to_next_on_post_save(sender, instance, raw, created, using, **kwargs):
    if (
        not raw
        and not kwargs.get("update_fields")
        and get_allow_skipped_appt_using().get(instance._meta.label_lower)
    ):
        skip_appt = SkipAppointments(instance)
        next_appointment_updated = skip_appt.update()
        allow_create_interim = getattr(instance, "allow_create_interim", False)
        if not next_appointment_updated and allow_create_interim:
            create_next_appointment_as_interim(
                next_appt_datetime=skip_appt.next_appt_datetime,
                appointment=skip_appt.appointment,
            )


@receiver(post_delete, weak=False, dispatch_uid="update_appointments_to_next_on_post_delete")
def update_appointments_to_next_on_post_delete(sender, instance, using, **kwargs):
    if get_allow_skipped_appt_using().get(instance._meta.label_lower):
        SkipAppointments(instance).reset_appointments()


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_appointment_from_nextappointment_post_save",
)
def update_appointment_from_nextappointment_post_save(
    sender, instance, raw, created, using, **kwargs
):
    if (
        not raw
        and not kwargs.get("update_fields")
        and isinstance(instance, (NextAppointmentCrfModelMixin,))
        and instance.offschedule_today == NO
    ):
        appointment = Appointment.objects.get(pk=instance.related_visit.appointment.next.id)
        appointment.appt_datetime = instance.appt_datetime
        appointment.save()



from django.db import models
from django_audit_fields.models import AuditModelMixin

from .url_model_mixin import UrlModelMixin

edc_default_permissions = ("add", "change", "delete", "view", "export", "import")


class BaseModel(UrlModelMixin, AuditModelMixin, models.Model):
    """Base model class for all EDC models. Adds created and modified'
    values for user, date and hostname (computer).
    """

    objects = models.Manager()

    @property
    def verbose_name(self):
        return self._meta.verbose_name

    class Meta(AuditModelMixin.Meta):
        abstract = True
        default_permissions = edc_default_permissions
        default_manager_name = "objects"
        indexes = AuditModelMixin.Meta.indexes



# Generated by Django 3.1.8 on 2021-04-23 11:51

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_visit_tracking", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="id",
            field=models.AutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="id",
            field=models.AutoField(primary_key=True, serialize=False),
        ),
    ]



from .action_item import DataManagerActionItem
from .data_dictionary import DataDictionary
from .data_query import DataQuery
from .edc_permissions import EdcPermissions
from .proxy_models import Issue
from .query_rule import (
    CrfDataDictionary,
    QueryRule,
    QueryRuleError,
    RequisitionDataDictionary,
    VisitDataDictionary,
    get_rule_handler_choices,
)
from .query_subject import QuerySubject
from .query_visit_schedule import QueryVisitSchedule
from .requisition_panel import RequisitionPanel
from .signals import update_query_text
from .user import DataManagerUser, QueryUser



import numpy as np
import pandas as pd
from django.apps import apps as django_apps
from django.db.models import Model
from django_pandas.io import read_frame

from edc_constants.constants import FEMALE, MALE

from ..utils import convert_dates_from_model


def get_subject_consent(
    model: str | None = None,
    model_cls: type[Model] | None = None,
    subject_identifiers: list[str] | None = None,
    extra_columns: list[str] | None = None,
    normalize: bool | None = None,
    localize: bool | None = None,
) -> pd.DataFrame:
    extra_columns = extra_columns or []
    normalize = True if normalize is None else normalize
    localize = True if localize is None else localize
    model_cls = model_cls or django_apps.get_model(model)
    value_cols = [
        "subject_identifier",
        "gender",
        "dob",
        "screening_identifier",
        "consent_datetime",
        "report_datetime",
        "created",
        "user_created",
        "version",
        "site",
    ]
    value_cols = list(set(value_cols + extra_columns))
    if subject_identifiers:
        qs_consent = model_cls.objects.values(*value_cols).filter(
            subject_identifier__in=subject_identifiers
        )
    else:
        qs_consent = model_cls.objects.values(*value_cols).all()
    df = read_frame(qs_consent, verbose=False)
    df = df.rename(columns={"site": "site_id"})
    df["gender"] = pd.Categorical(df["gender"], categories=[FEMALE, MALE], ordered=True)
    for col in ["dob", "consent_datetime", "report_datetime", "created"]:
        df[col] = df[col].apply(pd.to_datetime, errors="coerce")

    df = convert_dates_from_model(df, model_cls, normalize=normalize, localize=localize)
    if not df["consent_datetime"].empty:
        df["age_in_years"] = df["consent_datetime"].dt.year - df["dob"].dt.year
        df["age_in_years"] = df["age_in_years"].astype("int64")
    else:
        df["age_in_years"] = np.nan

    df = df.sort_values(by="subject_identifier")
    df = df.reset_index(drop=True)
    return df



from django.apps import apps as django_apps
from django.core.checks import CheckMessage, Error
from django.urls import NoReverseMatch

from edc_navbar import site_navbars


def edc_navbar_checks(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []

    for navbar_name, navbar in site_navbars.registry.items():
        for navbar_item in navbar.navbar_items:
            try:
                app_label, codename = navbar_item.codename.split(".")
            except ValueError:
                errors.append(
                    Error(
                        f"Invalid codename. Got '{navbar_item.codename}'."
                        "See {repr(navbar_item)}.",
                        id="edc_navbar.E001",
                    )
                )
            else:
                if app_label not in [a.name for a in django_apps.get_app_configs()]:
                    msg = (
                        f"Invalid app_label in codename. Expected format "
                        f"'<app_label>.<some_codename>'. Got {navbar_item.codename}. "
                        f"See {navbar_item!r}."
                    )
                    errors.append(Error(msg, id="edc_navbar.E002"))
            try:
                navbar_item.get_url(raise_exception=True)
            except NoReverseMatch:
                errors.append(
                    Error(
                        f"NoReverseMatch for url. Got '{navbar_item.real_url_name}'. "
                        f"See {navbar_item.name}.",
                        id="edc_navbar.E003",
                    )
                )

    return errors



from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django_audit_fields import audit_fieldset_tuple

from .admin_site import edc_label_admin
from .models import ZplLabelTemplates


@admin.register(ZplLabelTemplates, site=edc_label_admin)
class ZplLabelTemplatesAdmin(ModelAdmin):
    fieldsets = (
        [None, {"fields": ("name", "zpl_data")}],
        audit_fieldset_tuple,
    )

    search_fields: tuple[str, ...] = ("name",)



# Generated by Django 5.2.3 on 2025-07-08 12:21

import uuid

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import edc_pharmacy.models.stock.confirmation_at_site_item
import edc_sites.managers


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0083_auto_20250707_1820"),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RenameModel(
            old_name="HistoricalStockTransferConfirmationItem",
            new_name="HistoricalConfirmationAtSiteItem",
        ),
        migrations.RenameModel(
            old_name="StockTransferConfirmationItem",
            new_name="ConfirmationAtSiteItem",
        ),
        migrations.RenameModel(
            old_name="HistoricalStockTransferConfirmation",
            new_name="HistoricalConfirmationAtSite",
        ),
        migrations.RenameModel(
            old_name="StockTransferConfirmation",
            new_name="ConfirmationAtSite",
        ),
    ]



from contextlib import suppress

from edc_auth.site_auths import GroupAlreadyExists, RoleAlreadyExists, site_auths

from .auth_objects import (
    UNBLINDING_REQUESTORS,
    UNBLINDING_REQUESTORS_ROLE,
    UNBLINDING_REVIEWERS,
    UNBLINDING_REVIEWERS_ROLE,
    unblinding_requestors,
    unblinding_reviewers,
)


def update_site_auths() -> None:
    with suppress(GroupAlreadyExists):
        site_auths.add_group(*unblinding_requestors, name=UNBLINDING_REQUESTORS)
    with suppress(GroupAlreadyExists):
        site_auths.add_group(*unblinding_reviewers, name=UNBLINDING_REVIEWERS)
    with suppress(RoleAlreadyExists):
        site_auths.add_role(UNBLINDING_REQUESTORS, name=UNBLINDING_REQUESTORS_ROLE)
    with suppress(RoleAlreadyExists):
        site_auths.add_role(UNBLINDING_REVIEWERS, name=UNBLINDING_REVIEWERS_ROLE)


update_site_auths()



from typing import Any


class TemplateRequestContextError(Exception):
    pass


class TemplateRequestContextMixin:
    """Adds a method to get a specified template from the request.context_data.

    For example:
        def get_template_names(self):
            return [self.get_template_from_context(self.listboard_template)]
    """

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        """Adds template data to context."""
        kwargs.update(self.request.template_data)
        return super().get_context_data(**kwargs)

    def get_template_from_context(self, key=None):
        """Returns a template_name from request.context_data."""
        try:
            template_name = self.request.template_data[key]
        except KeyError as e:
            raise TemplateRequestContextError(
                f"Template name not defined in request context data. "
                f"Expected one of {list(self.request.template_data.keys())}. Got {e}. "
            )
        return template_name



# Generated by Django 5.1.2 on 2024-11-22 13:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0051_alter_historicalstocktransferconfirmationitem_options_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="formulation",
            name="imp",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="formulation",
            name="imp_description",
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name="historicalformulation",
            name="imp",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalformulation",
            name="imp_description",
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
    ]



from __future__ import annotations

from uuid import UUID

from celery import shared_task
from django.apps import apps as django_apps
from django.utils import timezone

from ..exceptions import InsufficientStockError, RepackError


@shared_task
def process_repack_request(repack_request_id: UUID | None = None, username: str | None = None):
    """Take from stock and fill container as new stock item."""
    repack_request_model_cls = django_apps.get_model("edc_pharmacy.repackrequest")
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    repack_request = repack_request_model_cls.objects.get(id=repack_request_id)
    repack_request.task_id = None
    repack_request.processed_qty = repack_request.processed_qty = (
        stock_model_cls.objects.filter(repack_request=repack_request).count()
    )
    repack_request.requested_qty = (
        repack_request.processed_qty
        if not repack_request.requested_qty
        else repack_request.requested_qty
    )
    number_to_process = repack_request.requested_qty - repack_request.processed_qty
    if not getattr(repack_request.from_stock, "confirmation", None):
        raise RepackError("Source stock item not confirmed")
    stock_model_cls = repack_request.from_stock.__class__
    for _ in range(0, int(number_to_process)):
        try:
            stock_model_cls.objects.create(
                receive_item=None,
                qty_in=1,
                qty_out=0,
                qty=1,
                from_stock=repack_request.from_stock,
                container=repack_request.container,
                location=repack_request.from_stock.location,
                repack_request=repack_request,
                lot=repack_request.from_stock.lot,
                user_created=username,
                created=timezone.now(),
            )
        except InsufficientStockError:
            break
    repack_request.processed_qty = stock_model_cls.objects.filter(
        repack_request=repack_request
    ).count()
    repack_request.user_modified = username
    repack_request.modified = timezone.now()
    repack_request.save(
        update_fields=[
            "requested_qty",
            "processed_qty",
            "task_id",
            "user_modified",
            "modified",
        ]
    )
    repack_request.refresh_from_db()


__all__ = ["process_repack_request"]



from django.contrib import admin, messages
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import User
from django.template.loader import render_to_string

from edc_dashboard.utils import select_edc_template
from edc_model_admin.mixins import TemplatesModelAdminMixin

from ..admin_site import edc_auth_admin
from ..forms import UserChangeForm
from ..send_new_credentials_to_user import send_new_credentials_to_user
from .list_filters import CountriesListFilter, SitesListFilter
from .user_profile_admin import UserProfileInline

admin.site.unregister(User)


def send_new_credentials_to_user_action(modeladmin, request, queryset):
    if request.user.has_perm("auth.change_user"):
        for obj in queryset:
            send_new_credentials_to_user(user=obj)
    else:
        messages.error(request, "You do not have permissions to run this action.")


send_new_credentials_to_user_action.short_description = "Reset password and email to user"


@admin.register(User, site=edc_auth_admin)
class UserAdmin(TemplatesModelAdminMixin, BaseUserAdmin):
    show_object_tools: bool = True

    form = UserChangeForm
    inlines = [UserProfileInline]
    actions = [send_new_credentials_to_user_action]

    list_display = (
        "username",
        "email",
        "first_name",
        "last_name",
        "role",
        "sites",
        "is_staff",
        "is_active",
        "multisite_viewer",
        "last_login",
    )

    search_fields = (
        "username",
        "first_name",
        "last_name",
        "email",
        "userprofile__roles__name",
    )

    list_filter = (
        "is_staff",
        "is_superuser",
        "is_active",
        "userprofile__is_multisite_viewer",
        CountriesListFilter,
        SitesListFilter,
        "userprofile__roles",
        "groups",
    )

    def get_inline_instances(self, request, obj=None):
        """Don't load inlines on add.

        Wait until after first post_save signal creates
        UserProfile.
        """
        inline_instances = []
        if obj is None:
            return inline_instances
        return super().get_inline_instances(request, obj=obj)

    @admin.display(description="Role")
    def role(self, obj=None) -> str:
        roles = []
        role_group_names = []
        for role in obj.userprofile.roles.all():
            roles.append(role)
            role_group_names.extend([grp.name for grp in role.groups.all().order_by("name")])
        extra_groups = [obj for obj in obj.groups.all() if obj.name not in role_group_names]
        context = dict(
            role_names=[role.display_name for role in roles],
            extra_group_names=[grp.name.replace("_", " ") for grp in extra_groups],
        )
        template_obj = select_edc_template("user_role_description.html", "edc_auth")
        return render_to_string(template_obj.template.name, context)

    @admin.display(description="Sites")
    def sites(self, obj=None) -> str:
        country_sites = {}
        for site in obj.userprofile.sites.all().order_by("siteprofile__country", "name"):
            country_name = site.siteprofile.country.replace("_", " ").title()
            site_name = site.name.replace("_", " ").title()
            try:
                country_sites[country_name].append(site_name)
            except KeyError:
                country_sites[country_name] = [site_name]

        context = dict(country_sites=country_sites)
        template_obj = select_edc_template("user_country_sites.html", "edc_auth")
        return render_to_string(template_obj.template.name, context)

    @admin.display(
        description="Multisite",
        boolean=True,
        ordering="userprofile__is_multisite_viewer",
    )
    def multisite_viewer(self, obj=None):
        return obj.userprofile.is_multisite_viewer



from django import forms

from edc_pylabels.models import LabelConfiguration
from edc_pylabels.site_label_configs import site_label_configs


class LabelConfigurationForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["name"].choices = [(k, k) for k in site_label_configs.all()]

    name = forms.ChoiceField(choices=[])

    def clean(self):
        cleaned_data = super().clean()
        if (
            "name" in cleaned_data
            and cleaned_data.get("name")
            and cleaned_data.get("name") not in site_label_configs.all()
        ):
            raise forms.ValidationError(
                "Invalid name. Name not registered with site_label_configs. "
                f"Expected one of {list(site_label_configs.all().keys())}"
            )
        return cleaned_data

    class Meta:
        model = LabelConfiguration
        fields = "__all__"



import re

from .base_form_validator import (
    OUT_OF_RANGE_ERROR,
    REQUIRED_ERROR,
    BaseFormValidator,
    InvalidModelFormFieldValidator,
)


class RangeFieldValidator(BaseFormValidator):
    def out_of_range_if(
        self,
        lower: int,
        upper: int,
        field: str = None,
        lower_inclusive: bool = None,
        upper_inclusive: bool = None,
        allow_none: bool = None,
    ) -> bool:
        r = re.compile(r"^\d*[.,]?\d*$")
        lower_inclusive = True if lower_inclusive is None else lower_inclusive
        upper_inclusive = True if upper_inclusive is None else upper_inclusive
        if not field:
            raise InvalidModelFormFieldValidator("The field attr cannot be None.")
        if self.cleaned_data:
            value = self.cleaned_data.get(field)
            if value is None and allow_none:
                pass
            elif value is None and not allow_none:
                self.raise_validation_error({field: "This field is required."}, REQUIRED_ERROR)
            elif r.match(str(value)):
                lower_op = "<" if not lower_inclusive else "<="
                upper_op = "<" if not upper_inclusive else "<="
                expression = f"{lower}{lower_op}{value}{upper_op}{upper}"
                if not eval(expression):  # nosec B307
                    message = {
                        field: f"This field is not within range. Expected {expression}."
                    }
                    self.raise_validation_error(message, OUT_OF_RANGE_ERROR)
        return False



from django.conf import settings
from django.core.checks import CheckMessage, Error
from django.core.management import color_style

style = color_style()

__all__ = ["check_for_edc_appconfig"]


def check_for_edc_appconfig(app_configs, **kwargs) -> list[CheckMessage]:
    """Check for edc_appconfig in INSTALLED_APPS.

    Register in edc_auth
    """
    errors = []
    if getattr(settings, "EDC_APPCONFIG_SYSTEM_CHECK_ENABLED", True):
        if "edc_appconfig.apps.AppConfig" not in settings.INSTALLED_APPS:
            errors.append(
                Error(
                    "edc_appconfig is not in INSTALLED_APPS.",
                    id="edc_appconfig.E001",
                )
            )
        if settings.INSTALLED_APPS[-1:][0] != "edc_appconfig.apps.AppConfig":
            errors.append(
                Error(
                    "edc_appconfig should be the last app in INSTALLED_APPS. "
                    f"Got {settings.INSTALLED_APPS[-1:]}",
                    id="edc_appconfig.E002",
                )
            )
    return errors



from django.contrib import admin

from ..admin_site import edc_export_admin
from ..models import FileHistory


@admin.register(FileHistory, site=edc_export_admin)
class FileHistoryAdmin(admin.ModelAdmin):
    date_hierarchy = "sent_datetime"
    list_display = (
        "model",
        "exit_status",
        "closed",
        "closed_datetime",
        "exported",
        "exported_datetime",
        "sent",
        "sent_datetime",
        "received",
        "received_datetime",
    )
    list_filter = (
        "exit_status",
        "model",
        "closed",
        "sent",
        "received",
        "exported",
        "closed_datetime",
        "exported_datetime",
        "sent_datetime",
        "received_datetime",
        "user_created",
    )
    search_fields = ("export_uuid_list", "pk_list")



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_email_contacts(key) -> dict:
    email_contacts = getattr(settings, "EMAIL_CONTACTS", {})
    if get_email_enabled() and key not in email_contacts:
        raise ImproperlyConfigured(
            f"Key not found. See settings.EMAIL_CONTACTS. Got key=`{key}`."
        )
    return email_contacts.get(key)


def get_email_enabled() -> bool:
    return getattr(settings, "EMAIL_ENABLED", False)



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse

from ...admin_site import edc_pharmacy_admin
from ...models import AllocationProxy
from .allocation_admin import AllocationAdmin


@admin.register(AllocationProxy, site=edc_pharmacy_admin)
class AllocationProxyAdmin(AllocationAdmin):

    @admin.display(description="Assignment")
    def assignment(self, obj):
        return None

    @admin.display(description="Stock #")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stockproxy_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=f"{obj.stock.code}", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .filter(
                stock__confirmation__isnull=False,
                stock__allocation__isnull=False,
                stock__container__may_request_as=True,
            )
        )



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from .constants import CLOSED_TIMEPOINT
from .timepoint import TimepointClosed


class TimepointLookup:
    """A class to be set as an attribute of a model
    controlled by a timepoint model.

    For example, a SubjectVisit model or a CrfModel. Verify that the
    default class attributes work for your model and timepoint model.
    If not, create a child class as has been done in the tests.

    Declare as:
        class CrfModel(...):

            timepoint_lookup_cls = TimepointLookup

            ...

    """

    timepoint_model = "edc_appointment.appointment"
    timepoint_related_model_lookup = "subject_visit__appointment"

    def __init__(self, timepoint_model=None, timepoint_related_model_lookup=None):
        self.timepoint_model = timepoint_model or self.timepoint_model
        self.timepoint_model_lookup = (
            timepoint_related_model_lookup or self.timepoint_related_model_lookup
        )

    def __str__(self):
        return self.timepoint_model

    @property
    def timepoint_model_cls(self):
        app_config = django_apps.get_app_config("edc_timepoint")
        return app_config.timepoints.get_model(self.timepoint_model)

    def raise_if_closed(self, model_obj=None):
        try:
            timepoint_obj = model_obj.__class__.objects.get(
                **{
                    f"{self.timepoint_related_model_lookup}"
                    "__timepoint_status": CLOSED_TIMEPOINT
                }
            )
        except ObjectDoesNotExist:
            pass
        else:
            raise TimepointClosed(f"Timepoint is closed for {model_obj}. See {timepoint_obj}.")



from edc_list_data.model_mixins import ListModelMixin


class HealthFacilityTypes(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Health Facility Type"
        verbose_name_plural = "Health Facility Types"



from django.db import models

from edc_model.models import BaseUuidModel, HistoricalRecords


class MolecularWeight(BaseUuidModel):

    label = models.CharField(max_length=25, unique=True)

    mw = models.FloatField(verbose_name="Molecular weight", default=0, help_text="in g/mol")

    history = HistoricalRecords()

    def __str__(self):
        return self.label

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Molecular weight"
        verbose_name_plural = "Molecular weights"



import re

from django.db import models
from django.db.models import UniqueConstraint
from django.db.models.deletion import PROTECT

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_search.model_mixins import SearchSlugManager, SearchSlugModelMixin

from ..model_mixins import VerifyModelMixin
from ..patterns import aliquot_pattern
from .box import Box


class BoxItemManager(SearchSlugManager, models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, position, identifier, box_identifier):
        return self.get(
            position=position, identifier=identifier, box_identifier=box_identifier
        )


class BoxItem(SearchSlugModelMixin, VerifyModelMixin, BaseUuidModel):
    box = models.ForeignKey(Box, on_delete=PROTECT)

    position = models.IntegerField()

    identifier = models.CharField(max_length=25)

    comment = models.CharField(max_length=25, default="", blank=True)

    objects = BoxItemManager()

    history = HistoricalRecords()

    def natural_key(self):
        return self.position, self.identifier, *self.box.natural_key()

    natural_key.dependencies = ("edc_lab.box", "edc_lab.boxtype", "sites.Site")

    @property
    def human_readable_identifier(self):
        """Returns a human-readable identifier"""
        if self.identifier:
            x = self.identifier
            if re.match(aliquot_pattern, self.identifier):
                return f"{x[0:3]}-{x[3:6]}-{x[6:10]}-{x[10:14]}-{x[14:18]}"
        return self.identifier

    def get_slugs(self):
        return self.identifier, self.human_readable_identifier

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Box Item"
        constraints = (
            UniqueConstraint(
                fields=["box", "position"], name="%(app_label)s_%(class)s_box_pos_uniq"
            ),
            UniqueConstraint(
                fields=["box", "identifier"],
                name="%(app_label)s_%(class)s_box_ide_uniq",
            ),
        )



inline_hide_object_original_name = True


class InlineHideOriginalObjectNameMixin:
    """Declare with main ModelAdmin class, not the InlineModelAdmin class.

    Only works if declared with the TabularInlineMixin.
    """

    inline_hide_object_original_name = True

    def get_formset(self, request, obj=None, **kwargs):
        formset = super().get_formset(request, obj=None, **kwargs)
        formset.insert_before_fieldset = self.insert_before_fieldset
        return formset

    def add_view(self, request, form_url: str = "", extra_context: dict = None):
        extra_context = extra_context or {}
        extra_context.update(
            inline_hide_object_original_name=self.inline_hide_object_original_name
        )
        return super().add_view(request, form_url=form_url, extra_context=extra_context)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context.update(
            inline_hide_object_original_name=self.inline_hide_object_original_name
        )
        return super().change_view(
            request, object_id, form_url=form_url, extra_context=extra_context
        )



from __future__ import annotations

import contextlib
import re
from typing import TYPE_CHECKING, Any

from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.urls import NoReverseMatch, reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_dashboard.url_names import InvalidDashboardUrlName, url_names

from .constants import ELIGIBLE, NOT_ELIGIBLE
from .exceptions import InvalidScreeningIdentifierFormat

if TYPE_CHECKING:
    from .model_mixins import EligibilityModelMixin, ScreeningModelMixin


def get_subject_screening_app_label() -> str:
    return get_subject_screening_model().split(".")[0]


def get_subject_screening_model() -> str:
    return settings.SUBJECT_SCREENING_MODEL


def get_subject_screening_model_cls() -> Any:
    return django_apps.get_model(get_subject_screening_model())


def format_reasons_ineligible(*str_values: str | None, delimiter: str | None = None) -> str:
    reasons = None
    delimiter = delimiter or "|"
    str_values = str_values or []
    str_values = tuple(x for x in str_values if x)
    if str_values:
        reasons = format_html(
            "{}",
            mark_safe(delimiter.join(str_values)),  # noqa: S308
        )
    return reasons


def eligibility_display_label(eligible) -> str:
    return ELIGIBLE.upper() if eligible else NOT_ELIGIBLE


def validate_screening_identifier_format_or_raise(
    screening_identifier: str,
    pattern: str | None = None,
    exception_cls: Exception | None = None,
) -> None:
    """Validates the identifier pattern or raises."""
    pattern = pattern or r"^[A-Z0-9]+$"
    if not screening_identifier or not re.match(pattern, screening_identifier or ""):
        raise (exception_cls or InvalidScreeningIdentifierFormat)(
            f"Invalid screening identifier. Got `{screening_identifier}`."
        )


def get_subject_screening_or_raise(
    screening_identifier: str, is_modelform: bool | None = None
) -> ScreeningModelMixin:
    """Returns the subject_screening model instance or raises."""
    try:
        with transaction.atomic():
            subject_screening = get_subject_screening_model_cls().objects.get(
                screening_identifier=screening_identifier
            )
    except ObjectDoesNotExist as e:
        if is_modelform:
            raise forms.ValidationError("Not allowed. Screening form not found.") from e
        raise ObjectDoesNotExist(
            f"{e} screening_identifier={screening_identifier}. Perhaps catch this in the form."
        ) from e
    return subject_screening


def is_eligible_or_raise(
    screening_identifier: str | None = None,
    subject_screening: EligibilityModelMixin | None = None,
    url_name: str | None = None,
) -> None:
    """Raise a ValidationError if subject_screening.eligible is False.

    * this func is NOT for the SubjectScreening form;
    * Adds a URl linking to the subject screening form in the validation message;
    * Default url name is `screening_listboard_url`.
    """
    if screening_identifier and subject_screening:
        raise TypeError(
            f"Expected one value not both. Got screening_identifier={screening_identifier} "
            f"and subject_screening={subject_screening}."
        )
    subject_screening = subject_screening or get_subject_screening_or_raise(
        screening_identifier, is_modelform=True
    )

    url_name = url_name or "screening_listboard_url"
    with contextlib.suppress(InvalidDashboardUrlName):
        url_name = url_names.get(url_name)

    if not subject_screening.eligible:
        try:
            url = reverse(
                url_name,
                kwargs={"screening_identifier": subject_screening.screening_identifier},
            )
        except NoReverseMatch:
            try:
                url = reverse(url_name)
            except NoReverseMatch:
                url = None
            if url and url_name.endswith("changelist"):
                url = f"{url}?q={subject_screening.screening_identifier}"
        if not url:
            msg = format_html(
                "{}",
                "Not allowed. Subject is not eligible. "
                f"Got {subject_screening.screening_identifier}",
            )
        else:
            msg = format_html(
                'Not allowed. Subject is not eligible. See subject <A href="{}">{}</A>',
                mark_safe(url),  # noqa: S308
                subject_screening.screening_identifier,
            )
        raise forms.ValidationError(msg)



from django.contrib import messages

from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab.labels import AliquotLabel

from ...view_mixins import ProcessRequisitionViewMixin
from .action_view import ActionView


class ProcessView(EdcViewMixin, ProcessRequisitionViewMixin, ActionView):
    post_action_url = "process_listboard_url"
    valid_form_actions = ["process"]
    action_name = "process"
    label_cls = AliquotLabel

    def process_form_action(self, request=None):
        if self.action == "process":
            if not self.selected_items:
                message = "Nothing to do. No items have been selected."
                messages.warning(request, message)
            else:
                self.process(request)



from django.urls.conf import path, re_path

from .admin_site import edc_export_admin
from .views import ExportModelsView, ExportSelectedModelsView, HomeView

app_name = "edc_export"

urlpatterns = [
    path("admin/", edc_export_admin.urls),
    path(
        "selected_models/",
        ExportSelectedModelsView.as_view(),
        name="export_selected_models_url",
    ),
    re_path(
        "models/(?P<action>cancel|confirm)/",
        ExportModelsView.as_view(),
        name="export_models_url",
    ),
    re_path("models/", ExportModelsView.as_view(), name="export_models_url"),
    path("", HomeView.as_view(), name="home_url"),
]



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_registration",
            "0029_alter_historicalregisteredsubject_device_created_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="registeredsubject",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "permissions": (
                    ("display_firstname", "Can display first name"),
                    ("display_lastname", "Can display last name"),
                    ("display_dob", "Can display DOB"),
                    ("display_identity", "Can display identity number"),
                    ("display_initials", "Can display initials"),
                ),
                "verbose_name": "Registered Subject",
                "verbose_name_plural": "Registered Subjects",
            },
        ),
        migrations.RemoveIndex(
            model_name="registeredsubject",
            name="edc_registr_identit_eb36e0_idx",
        ),
        migrations.AlterUniqueTogether(
            name="registeredsubject",
            unique_together=set(),
        ),
        migrations.AddField(
            model_name="historicalregisteredsubject",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalregisteredsubject",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="registeredsubject",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="registeredsubject",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddIndex(
            model_name="registeredsubject",
            index=models.Index(
                fields=["modified", "created"], name="edc_registr_modifie_d58ae9_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="registeredsubject",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_registr_user_mo_079b86_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="registeredsubject",
            index=models.Index(
                fields=["subject_identifier", "identity", "screening_identifier"],
                name="edc_registr_subject_6bb1e5_idx",
            ),
        ),
        migrations.AddConstraint(
            model_name="registeredsubject",
            constraint=models.UniqueConstraint(
                fields=("first_name", "dob", "initials", "additional_key"),
                name="edc_registration_registeredsubject_first_name_uniq",
            ),
        ),
    ]



from typing import cast

from django.db import models

from .action import Action
from .action_item_notification import ActionItemNotification


class ActionWithNotification(Action):
    """A class mixin for the Action class that adds
    support for notifications.
    """

    notification_email_to: list[str] = None
    notification_display_name: str = None
    notification_fields: list[str] = None
    notification_super_cls: type[ActionItemNotification] = ActionItemNotification
    notify_on_changed_reference_obj: models.Model = True
    notify_on_close: bool = False
    notify_on_new: bool = False
    notify_on_new_and_no_reference_obj: bool = True
    notify_on_open: bool = False

    @classmethod
    def notification_cls(cls) -> type[ActionItemNotification]:
        """Returns a subclass of ActionItemModelNotification."""
        return cast(
            type[ActionItemNotification],
            type(
                f"{cls.__name__}Notification",
                (cls.notification_super_cls,),
                dict(
                    name=f"{cls.name}-notification",
                    notification_action_name=cls.name,
                    display_name=(
                        cls.notification_display_name or f"{cls.display_name} Notification"
                    ),
                    email_to=cls.notification_email_to,
                    notification_fields=cls.notification_fields,
                    model=cls.get_reference_model(),
                    notify_on_new_and_no_reference_obj=cls.notify_on_new_and_no_reference_obj,
                    notify_on_new=cls.notify_on_new,
                    notify_on_open=cls.notify_on_open,
                    notify_on_close=cls.notify_on_close,
                    notify_on_changed_reference_obj=cls.notify_on_changed_reference_obj,
                ),
            ),
        )



ACTION_ITEM_COLUMNS = [
    "action_identifier",
    "action_item_id",
    "action_item_reason",
    "parent_action_identifier",
    "parent_action_item_id",
    "related_action_identifier",
    "related_action_item_id",
]

SYSTEM_COLUMNS = [
    "created",
    "modified",
    "user_created",
    "user_modified",
    "hostname_created",
    "hostname_modified",
    "device_created",
    "device_modified",
    "locale_created",
    "locale_modified",
    "revision",
]

numeric_datatypes = [
    "DecimalField",
    "IntegerField",
    "BigIntegerField",
    "FloatField",
    "PositiveBigIntegerField",
    "PositiveIntegerField",
    "PositiveSmallIntegerField",
    "SmallIntegerField",
]

date_datatypes = ["DateTimeField", "DateField"]

timedelta_datatypes = ["DurationField"]



# Generated by Django 5.2.4 on 2025-09-10 03:24

import uuid

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_reportable", "0005_historicalnormaldata_auto_created_and_more"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name="gradingdata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="gradingexception",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.CreateModel(
            name="HistoricalMolecularWeight",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("label", models.CharField(db_index=True, max_length=25)),
                (
                    "mw",
                    models.FloatField(
                        default=0, help_text="in g/mol", verbose_name="Molecular weight"
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Molecular weight",
                "verbose_name_plural": "historical Molecular weights",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="MolecularWeight",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("label", models.CharField(max_length=25, unique=True)),
                (
                    "mw",
                    models.FloatField(
                        default=0, help_text="in g/mol", verbose_name="Molecular weight"
                    ),
                ),
            ],
            options={
                "verbose_name": "Molecular weight",
                "verbose_name_plural": "Molecular weights",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "default_manager_name": "objects",
                "indexes": [
                    models.Index(
                        fields=["modified", "created"],
                        name="edc_reporta_modifie_7e6235_idx",
                    ),
                    models.Index(
                        fields=["user_modified", "user_created"],
                        name="edc_reporta_user_mo_e8f92b_idx",
                    ),
                ],
            },
        ),
    ]



from edc_action_item.action_with_notification import ActionWithNotification
from edc_adverse_event.constants import (
    AE_FOLLOWUP_ACTION,
    AE_INITIAL_ACTION,
    AE_TMG_ACTION,
)
from edc_constants.constants import CLOSED, HIGH_PRIORITY

from ..constants import ADVERSE_EVENT_ADMIN_SITE, ADVERSE_EVENT_APP_LABEL


class AeTmgAction(ActionWithNotification):
    name = AE_TMG_ACTION
    display_name = "TMG AE Report pending"
    notification_display_name = "TMG AE Report"
    parent_action_names = (AE_INITIAL_ACTION, AE_FOLLOWUP_ACTION, AE_TMG_ACTION)
    reference_model = f"{ADVERSE_EVENT_APP_LABEL}.aetmg"
    related_reference_model = f"{ADVERSE_EVENT_APP_LABEL}.aeinitial"
    related_reference_fk_attr = "ae_initial"
    create_by_user = False
    color_style = "info"
    show_link_to_changelist = True
    admin_site_name = ADVERSE_EVENT_ADMIN_SITE
    instructions = "This report is to be completed by the TMG only."
    priority = HIGH_PRIORITY

    def close_action_item_on_save(self):
        return self.reference_obj.report_status == CLOSED



from edc_action_item.action import Action
from edc_action_item.action_with_notification import ActionWithNotification
from edc_constants.constants import HIGH_PRIORITY, YES
from edc_ltfu.constants import LTFU_ACTION

from .constants import VISIT_MISSED_ACTION


class VisitMissedAction(ActionWithNotification):
    name = VISIT_MISSED_ACTION
    display_name = "Submit Missed Visit"
    notification_display_name = " Submit Missed Visit"
    parent_action_names = ()
    show_link_to_changelist = True
    priority = HIGH_PRIORITY
    loss_to_followup_action_name = LTFU_ACTION

    reference_model = None  # "inte_subject.subjectvisitmissed"
    admin_site_name = None  # "inte_prn_admin"

    def get_loss_to_followup_action_name(self) -> str:
        return self.loss_to_followup_action_name

    def is_ltfu(self) -> bool:
        return self.reference_obj.ltfu == YES

    def get_next_actions(self) -> list[Action]:
        next_actions: list[Action] = []
        return self.append_to_next_if_required(
            next_actions=next_actions,
            action_name=self.get_loss_to_followup_action_name(),
            required=self.is_ltfu(),
        )


class MissedVisitAction(VisitMissedAction):
    pass



from django import forms

from edc_form_validators import INVALID_ERROR
from edc_reportable import ConversionNotHandled
from edc_vitals.calculators import CalculatorError

from ..calculators import EgfrCalculatorError, EgfrCkdEpi, EgfrCockcroftGault


class EgfrCkdEpiFormValidatorMixin:
    def validate_egfr(
        self,
        gender: str = None,
        age_in_years: int = None,
        weight_in_kgs: float = None,
        ethnicity: str = None,
        baseline_egfr_value: float | None = None,
    ):
        opts = dict(
            gender=gender,
            age_in_years=age_in_years,
            weight=weight_in_kgs,
            ethnicity=ethnicity,
            creatinine_value=self.cleaned_data.get("creatinine_value"),
            creatinine_units=self.cleaned_data.get("creatinine_units"),
            baseline_egfr_value=baseline_egfr_value,
        )
        try:
            value = EgfrCkdEpi(**opts).value
        except (EgfrCalculatorError, CalculatorError, ConversionNotHandled) as e:
            raise forms.ValidationError(e)
        return value


class EgfrCockcroftGaultFormValidatorMixin:
    def validate_egfr(
        self,
        *,
        gender: str,
        age_in_years: int,
        weight_in_kgs: float,
        ethnicity: str,
    ):
        opts = dict(
            gender=gender,
            age_in_years=age_in_years,
            weight=weight_in_kgs,
            ethnicity=ethnicity,
            creatinine_value=self.cleaned_data.get("creatinine_value"),
            creatinine_units=self.cleaned_data.get("creatinine_units"),
        )
        try:
            value = EgfrCockcroftGault(**opts).value
        except (EgfrCalculatorError, CalculatorError, ConversionNotHandled) as e:
            self.raise_validation_error({"__all__": str(e)}, INVALID_ERROR)
        return value



from edc_form_validators import FormValidator

INVALID_CLINIC_DAY = "INVALID_CLINIC_DAY"


class HealthFacilityFormValidator(FormValidator):
    def clean(self):
        mon = self.cleaned_data.get("mon")
        tue = self.cleaned_data.get("tue")
        wed = self.cleaned_data.get("wed")
        thu = self.cleaned_data.get("thu")
        fri = self.cleaned_data.get("fri")
        sat = self.cleaned_data.get("sat")
        sun = self.cleaned_data.get("sun")
        if not any([mon, tue, wed, thu, fri, sat, sun]):
            self.raise_validation_error(
                "Select at least one clinic day.", error_code=INVALID_CLINIC_DAY
            )



from .field_attrs import get_field_attrs_for_reportable, get_field_attrs_for_utestid
from .reportable_result_model_mixin_factory import reportable_result_model_mixin_factory
from .result_model_mixin_factory import result_model_mixin_factory

__all__ = [
    "get_field_attrs_for_reportable",
    "get_field_attrs_for_utestid",
    "reportable_result_model_mixin_factory",
    "result_model_mixin_factory",
]



# Generated by Django 4.2.7 on 2023-12-05 19:50

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0033_datadictionary_locale_created_and_more"),
        ("edc_form_runners", "0002_alter_issue_options_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Issue",
            fields=[],
            options={
                "verbose_name": "Form validation issue",
                "verbose_name_plural": "Form validation issues",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("edc_form_runners.issue",),
        ),
    ]



from __future__ import annotations

import re
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from django import forms
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import CharField, DateField, DateTimeField, Model
from django.urls import reverse

from .constants import REPORT_DATETIME_FIELD_NAME

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitTrackingCrfModelMixin

    from .models import BaseUuidModel

    class CrfLikeModel(VisitTrackingCrfModelMixin, BaseUuidModel):
        pass


dh_pattern = r"^(([0-9]{1,3}d)?((2[0-3]|[01]?[0-9])h)?)$"
# hm_pattern = r"^(([0-9]{1,2}h)?([0-9]{1,2}m)?)$"
hm_pattern = r"^(([0-9]{1,2}h)?([0-5]?[0-9]m)?)$"
ymd_pattern = (
    r"^([0-9]{1,3}y([0-1]?[0-2]m)?([0-9]m)?)$|^([0-1]?"
    r"[0-2]m)$|^([0-9]m)$|^([1-3]?[1-9]d)$|^(0d)$"
)


class InvalidFormat(Exception):
    pass


class InvalidFieldName(Exception):
    pass


def get_report_datetime_field_name():
    return getattr(settings, "EDC_MODEL_REPORT_DATETIME_FIELD_NAME", "report_datetime")


def estimated_date_from_ago(
    cleaned_data: dict = None,
    instance: Model = None,
    ago_field: str = None,
    reference_field: str = None,
    future: bool | None = None,
) -> date:
    """Wrapper function for `duration_to_date` typically called in
    modelform.clean() and model.save().

    Returns the estimated date using `duration_to_date` or None.

    Will raise an exception if the string cannot be interpreted.
    """

    estimated_date = None
    data = instance or cleaned_data
    reference_field = reference_field or REPORT_DATETIME_FIELD_NAME
    raise_on_invalid_field_name(data, ago_field)
    raise_on_invalid_field_name(data, reference_field)
    try:
        ago_str = data.get(ago_field)
        reference_date = data.get(reference_field)
    except AttributeError:
        ago_str = getattr(data, ago_field, None)
        reference_date = getattr(data, reference_field, None)
    try:
        reference_date = reference_date.date()
    except AttributeError:
        pass
    if ago_str and reference_date:
        try:
            estimated_date = duration_to_date(ago_str, reference_date, future=future)
        except InvalidFormat as e:
            if cleaned_data:
                raise forms.ValidationError({ago_field: str(e)})
            raise
    return estimated_date


def duration_to_date(
    duration_text: str | CharField,
    reference_date: date | datetime | DateField | DateTimeField,
    future: bool | None = None,
) -> date:
    """Returns the estimated date from a well-formatted string
    relative to a reference date/datetime.

    Will raise an exception if the string cannot be interpreted.
    """
    duration_text = duration_text.replace(" ", "")
    if not re.match(ymd_pattern, duration_text):
        raise InvalidFormat(
            "Expected format `NNyNNm`. e.g: 5y6m, 15y 12m, 12m, 4y... "
            "You may also specify number of days alone, e.g. 7d, 0d... "
            f"Got {duration_text}"
        )
    future = False if future is None else future
    years = 0
    months = 0
    days = 0
    if "y" in duration_text:
        years_str, remaining = duration_text.split("y")
        years = int(years_str)
        if remaining and "m" in remaining:
            months = int(remaining.split("m")[0])
    elif "m" in duration_text:
        months_str = duration_text.split("m")[0]
        months = int(months_str)
    else:
        days_str = duration_text.split("d")[0]
        days = int(days_str)
    delta = relativedelta(years=years, months=months, days=days)
    if future:
        return reference_date + delta
    return reference_date - delta


def raise_on_invalid_field_name(data: dict | models.Model, attrname: str) -> None:
    if attrname is not None:
        try:
            data[attrname]
        except KeyError as e:
            raise InvalidFieldName(f"{e} Got {data}")
        except TypeError:
            try:
                getattr(data, attrname)
            except AttributeError as e:
                raise InvalidFieldName(f"{e} Got {data}")
    else:
        raise InvalidFieldName(f"Field name cannot be None. Got {data}")


def model_exists_or_raise(
    subject_visit,
    model_cls,
    singleton=None,
) -> CrfLikeModel:
    singleton = False if singleton is None else singleton
    if singleton:
        opts = {"subject_visit__subject_identifier": subject_visit.subject_identifier}
    else:
        opts = {"subject_visit": subject_visit}
    try:
        obj = model_cls.objects.get(**opts)
    except ObjectDoesNotExist:
        raise ObjectDoesNotExist(f"{model_cls._meta.verbose_name} does not exist.")
    return obj


def is_inline_model(instance):
    """See also, edc-crf inline model mixin."""
    try:
        instance._meta.crf_inline_parent
    except AttributeError:
        return False
    return True


def timedelta_from_duration_dh_field(
    data: dict | models.Model, duration_dh_field: str
) -> timedelta | None:
    """Wrapper function for `duration_dh_to_timedelta` typically called in
    modelform.clean() and model.save().

    Perhaps just use duration_dh_to_timedelta?

    Returns timedelta using `duration_dh_to_timedelta` or None.

    Will raise an exception if the string cannot be interpreted.
    """
    duration_timedelta = None
    is_form_data = False
    raise_on_invalid_field_name(data, duration_dh_field)
    try:
        duration_dh_str = data.get(duration_dh_field)
    except AttributeError:
        duration_dh_str = getattr(data, duration_dh_field, None)
    else:
        is_form_data = True

    if duration_dh_str:
        try:
            duration_timedelta = duration_dh_to_timedelta(duration_text=duration_dh_str)
        except InvalidFormat as e:
            if is_form_data:
                raise forms.ValidationError({duration_dh_field: str(e)})
            raise
    return duration_timedelta


def duration_dh_to_timedelta(duration_text: str | CharField) -> timedelta:
    """Returns timedelta from a well-formatted string
    (specified in days and/or hours).

    Will raise an exception if the string cannot be interpreted.
    """
    duration_text = duration_text.replace(" ", "")
    if not re.match(dh_pattern, duration_text):
        raise InvalidFormat(
            "Expected format is `DDdHHh`, `DDd` or `HHh`. "
            "For example 1d23h, 15d9h ... or 20d, or 5h ..."
            f"Got {duration_text}"
        )
    days = 0
    hours = 0
    if "d" in duration_text:
        days_str, remaining = duration_text.split("d")
        days = int(days_str)
        if remaining and "h" in remaining:
            hours = int(remaining.split("h")[0])
    else:
        hours_str = duration_text.split("h")[0]
        hours = int(hours_str)
    return timedelta(days=days, hours=hours)


def duration_hm_to_timedelta(duration_text: str | CharField) -> timedelta:
    """Returns timedelta from a well-formatted string
    (specified in hours and/or mins).

    Will raise an exception if the string cannot be interpreted.
    """
    duration_text = duration_text.replace(" ", "")
    if not re.match(hm_pattern, duration_text):
        raise InvalidFormat(
            "Expected format is `HHhMMm`, `HHh` or `MMm`. "
            "For example 12h30m, 5h9m ... or 20h, or 5m ..."
            f"Got {duration_text}"
        )
    hours = 0
    minutes = 0
    if "h" in duration_text:
        hour_str, remaining = duration_text.split("h")
        hours = int(hour_str)
        if remaining and "m" in remaining:
            minutes = int(remaining.split("m")[0])
    else:
        minutes_str = duration_text.split("m")[0]
        minutes = int(minutes_str)
    return timedelta(hours=hours, minutes=minutes)


def get_history_url(obj, admin_site: str = None) -> str:
    admin_site = admin_site or obj.admin_url_name.split(":")[0]
    app, model = obj._meta.app_label, obj._meta.model_name
    return reverse(f"{admin_site}:{app}_{model}_history", args=[str(obj.pk)])



# Generated by Django 5.2.3 on 2025-07-07 15:16

import uuid

import _socket
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0079_remove_historicalstock_allocated_and_more"),
        ("sites", "0002_alter_domain_unique"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalstock",
            name="allocated_datetime",
        ),
        migrations.RemoveField(
            model_name="historicalstock",
            name="dispensed",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="allocated_datetime",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="dispensed",
        ),
        migrations.AlterField(
            model_name="historicalstoragebinitem",
            name="stock",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                limit_choices_to={
                    "confirmed_at_site": True,
                    "dispenseitem__isnull": True,
                },
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.stock",
            ),
        ),
        migrations.AlterField(
            model_name="storagebinitem",
            name="stock",
            field=models.ForeignKey(
                limit_choices_to={
                    "confirmed_at_site": True,
                    "dispenseitem__isnull": True,
                },
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stock",
            ),
        ),
    ]



from django.core.management.base import BaseCommand

from ...metadata_rules import site_metadata_rules


class Command(BaseCommand):
    help = "Performs a `get_model` for each target models referenced"

    def handle(self, *args, **options):
        site_metadata_rules.validate()



from __future__ import annotations

from typing import TYPE_CHECKING

from django.utils.translation import gettext_lazy as _

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


class ModelAdminFormInstructionsMixin:
    """Add instructions to the add view context.

    Override the change_form.html to add instructions.

    Copy change_form.html from the apps templates/<app>/admin/ into
    /templates/admin/<your_app> or create a blank change_form.html in your
    /templates/admin/<your_app> folder and add this:

        {% extends "admin/change_form.html" %}
        {% block field_sets %}
            {% instructions %}
            {% additional_instructions %}
            {{ block.super }}
        {% endblock %}

    See also edc_admin_modify.py for the inclusion tags registered
    `instructions` and `additional_instructions` and the templates
    involved. These templates can be overridden just as django
    non-admin templates.
    """

    instructions = _(
        "Please complete the form below. "
        "Required questions are in bold. "
        "When all required questions are complete click SAVE "
        "or, if available, SAVE NEXT. Based on your responses, "
        "additional questions may be "
        "required or some answers may need to be corrected."
    )
    add_instructions = ()
    change_instructions = ()

    additional_instructions = ()
    add_additional_instructions = None
    change_additional_instructions = ()

    def get_add_instructions(
        self, extra_context: dict | None, request: WSGIRequest | None = None
    ):
        extra_context = extra_context or {}
        extra_context["instructions"] = self.add_instructions or self.instructions
        extra_context["additional_instructions"] = (
            self.add_additional_instructions or self.additional_instructions
        )
        return extra_context

    def get_change_instructions(
        self, extra_context: dict | None, request: WSGIRequest | None = None
    ):
        extra_context = extra_context or {}
        extra_context["instructions"] = self.change_instructions or self.instructions
        extra_context["additional_instructions"] = (
            self.change_additional_instructions or self.additional_instructions
        )
        return extra_context

    def add_view(self, request: WSGIRequest, form_url: str = "", extra_context: dict = None):
        extra_context = self.get_add_instructions(extra_context, request=request)
        return super().add_view(request, form_url=form_url, extra_context=extra_context)

    def change_view(self, request: WSGIRequest, object_id, form_url="", extra_context=None):
        extra_context = self.get_change_instructions(extra_context, request=request)
        return super().change_view(
            request, object_id, form_url=form_url, extra_context=extra_context
        )



from django.contrib.admin import ModelAdmin
from django.contrib.admin.decorators import register

from ..admin_site import edc_pharmacy_admin
from ..models import SiteProxy, VisitSchedule


@register(VisitSchedule, site=edc_pharmacy_admin)
class VisitScheduleAdmin(ModelAdmin):
    ordering = ("visit_schedule_name", "schedule_name", "visit_code")
    search_fields = ("visit_code", "visit_title")


@register(SiteProxy, site=edc_pharmacy_admin)
class SiteProxyAdmin(ModelAdmin):
    ordering = ("name",)
    search_fields = ["id", "name"]



from __future__ import annotations

import re
from typing import Any

from django.contrib.messages import WARNING
from django.db.models import Q
from django.db.models.constants import LOOKUP_SEP
from django.utils.html import escape
from django.utils.translation import gettext as _

from edc_model_admin.utils import add_to_messages_once


class SearchListboardMixin:
    search_fields = ["slug"]

    default_querystring_attrs: str = "q"
    alternate_search_attr: str = "subject_identifier"
    default_lookup = "icontains"
    operators: list[str] = [
        "exact",
        "iexact",
        "contains",
        "icontains",
        "regex",
        "iregex",
        "gt",
        "gte",
        "lt",
        "lte",
        "startswith",
        "endswith",
        "istartswith",
        "iendswith",
    ]

    def __init__(self, **kwargs):
        self._search_term = None
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(search_term=self.search_term)
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        """Override to add conditional logic to filter on search term."""
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        if self.search_term and not re.match(r"^[A-Za-z0-9\-]+?$", self.search_term):
            add_to_messages_once(
                request=request,
                level=WARNING,
                message=_("Invalid search term. May only include letters, numbers and '-'."),
            )
        elif self.search_term:
            for field, lookup in self.get_field_lookups():
                q_object |= Q(**{f"{field}__{lookup}": self.search_term})
        return q_object, options

    @property
    def raw_search_term(self) -> str | None:
        raw_search_term = self.request.GET.get(self.default_querystring_attrs)
        if not raw_search_term:
            raw_search_term = self.kwargs.get(self.alternate_search_attr)
        return raw_search_term

    @property
    def search_term(self) -> str | None:
        if not self._search_term:
            if search_term := self.raw_search_term:
                self._search_term = escape(search_term).strip()
        return self._search_term

    def get_search_fields(self) -> list[str]:
        """Override to add additional search fields"""
        return self.search_fields

    def get_field_lookups(self) -> list[tuple[str, str]]:
        fields = []

        for field in self.get_search_fields():
            nlookups = field.split(LOOKUP_SEP)
            if nlookups[-1:][0] in self.operators:
                fld = LOOKUP_SEP.join(nlookups[:-1])
                lookup = nlookups[-1:][0]
            else:
                fld, lookup = LOOKUP_SEP.join(nlookups), self.default_lookup
            fields.append((fld, lookup))
        return fields



from django.urls.conf import path

from edc_dashboard.url_names import url_names
from edc_metadata.views import RefreshMetadataActionsView

from .views import (
    RefreshAppointmentsView,
    RequisitionPrintActionsView,
    RequisitionVerifyActionsView,
)

app_name = "edc_subject_dashboard"


urlpatterns = [
    path(
        "requisition_print_actions/",
        RequisitionPrintActionsView.as_view(),
        name="requisition_print_actions_url",
    ),
    path(
        "requisition_verify_actions/",
        RequisitionVerifyActionsView.as_view(),
        name="requisition_verify_actions_url",
    ),
    path(
        "refresh_subject_dashboard/<str:visit_schedule_name>/<str:schedule_name>"
        "/<str:related_visit_id>/",
        RefreshMetadataActionsView.as_view(),
        name="refresh_metadata_actions_url",
    ),
    path(
        "refresh_appointments/<str:visit_schedule_name>/<str:schedule_name>/"
        "<str:subject_identifier>/",
        RefreshAppointmentsView.as_view(),
        name="refresh_appointments_url",
    ),
]

url_names.register(url="requisition_print_actions_url", namespace=app_name)
url_names.register(url="requisition_verify_actions_url", namespace=app_name)
url_names.register(url="refresh_metadata_actions_url", namespace=app_name)



from __future__ import annotations

import re
from collections.abc import Iterable
from pathlib import Path

from django import forms
from django.conf import settings
from django.utils.html import format_html

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .constants import EXPORT_PII
from .exceptions import ExporterExportFolder


def get_export_folder() -> Path:
    if path := getattr(settings, "EDC_EXPORT_EXPORT_FOLDER", None):
        return Path(path).expanduser()
    return Path(settings.MEDIA_ROOT) / "data_folder" / "export"


def get_base_dir() -> Path:
    """Returns the base_dir used by, for example,
    shutil.make_archive.

    This is the short protocol name in lower case
    """
    base_dir: str = ResearchProtocolConfig().protocol_lower_name
    if len(base_dir) > 25:
        raise ExporterExportFolder(
            f"Invalid basedir, too long. Using `protocol_lower_name`. Got `{base_dir}`."
        )
    if not re.match(r"^[a-z0-9]+(?:_[a-z0-9]+)*$", base_dir):
        raise ExporterExportFolder(
            "Invalid base_dir, invalid characters. Using `protocol_lower_name`. "
            f"Got `{base_dir}`."
        )
    return Path(base_dir)


def get_upload_folder() -> Path:
    if path := getattr(settings, "EDC_EXPORT_UPLOAD_FOLDER", None):
        return Path(path).expanduser()
    return Path(settings.MEDIA_ROOT) / "data_folder" / "upload"


def get_export_pii_users() -> list[str]:
    return getattr(settings, "EDC_EXPORT_EXPORT_PII_USERS", [])


def raise_if_prohibited_from_export_pii_group(username: str, groups: Iterable) -> None:
    """A user form validation to prevent adding an unlisted
    user to the EXPORT_PII group.

    See also edc_auth's UserForm.
    """
    if EXPORT_PII in [grp.name for grp in groups] and username not in get_export_pii_users():
        raise forms.ValidationError(
            {
                "groups": format_html(
                    "This user is not allowed to export PII data. You may not add "
                    "this user to the <U>{text}</U> group.",
                    text="EXPORT_PII",
                )
            }
        )



from __future__ import annotations

from uuid import uuid4

from django.contrib import admin, messages
from django.db.models import QuerySet
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext

from ...models import Receive, RepackRequest, Stock


@admin.display(description="Confirm repacked and labeled stock")
def confirm_repacked_stock_action(modeladmin, request, queryset: QuerySet[RepackRequest]):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        return confirm_stock_from_queryset(modeladmin, request, queryset)
    return None


@admin.display(description="Confirm received and labeled stock")
def confirm_received_stock_action(modeladmin, request, queryset: QuerySet[RepackRequest]):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        return confirm_stock_from_queryset(modeladmin, request, queryset)
    return None


@admin.display(description="Confirm labeled stock")
def confirm_stock_from_queryset(
    modeladmin, request, queryset: QuerySet[RepackRequest | Receive | Stock]
):
    stock_codes = []
    pk = None
    identifier = None
    model_name = None
    word = "confirmed"
    if queryset.count() == 1:
        if queryset.model == RepackRequest:
            stock_qs = Stock.objects.filter(
                repack_request__pk__in=[obj.pk for obj in queryset]
            )
            stock_codes = stock_qs.values_list("code", flat=True)
            pk = queryset.first().pk
            identifier = queryset.first().repack_identifier
            model_name = queryset.model._meta.verbose_name
        elif queryset.model == Receive:
            stock_qs = Stock.objects.filter(
                receive_item__receive__pk__in=[obj.pk for obj in queryset]
            )
            stock_codes = stock_qs.values_list("code", flat=True)
            pk = queryset.first().pk
            identifier = queryset.first().receive_identifier
            model_name = queryset.model._meta.verbose_name
    elif queryset.count() > 1:
        stock_codes = queryset.values_list("code", flat=True)

    if queryset.count() >= 1:
        session_uuid = str(uuid4())
        request.session[session_uuid] = {
            "stock_codes": [str(o) for o in stock_codes],
            "source_pk": pk,
            "source_identifier": identifier,
            "source_label_lower": queryset.model._meta.label_lower,
            "source_model_name": model_name,
            "transaction_word": word,
        }
        url = reverse(
            "edc_pharmacy:confirm_stock_from_queryset_url",
            kwargs={"session_uuid": session_uuid},
        )
        return HttpResponseRedirect(url)
    return None



from __future__ import annotations

from typing import TYPE_CHECKING

from .constants import (
    MEAN_95CI,
    MEAN_RANGE,
    MEAN_SD,
    MEDIAN_IQR,
    MEDIAN_RANGE,
    N_MEAN,
    N_ONLY,
    N_WITH_COL_PROP,
    N_WITH_ROW_PROP,
    STATISTICS,
)

if TYPE_CHECKING:
    from .row import RowStatistics


class StylerError(Exception):
    pass


class Styler:
    """A class to format statistics per the format label given."""

    def __init__(
        self,
        style: str = None,
        statistics: RowStatistics = None,
        places: int | None = None,
    ):
        self.style = style
        self.row = statistics
        self.places = places if places is not None else 2
        if style not in STATISTICS:
            raise StylerError(f"Unknown style. Got `{style}`.")

    @property
    def value(self):
        """Make sure values are numerics first!

        For example, when preparing the dataframe convert values to
        numerics:
            df[cols] = df[cols].apply(pd.to_numeric)
        """
        col_value = "no style"
        if self.style == N_WITH_ROW_PROP:
            col_value = (
                f"{self.row.count} ("
                f"{round(self.row.rowprop * 100, self.places):.{self.places}f}%)"
            )
        elif self.style == N_ONLY:
            col_value = f"{self.row.count}"
        elif self.style == N_WITH_COL_PROP:
            col_value = (
                f"{self.row.count:.{self.places}f} "
                f"({round(self.row.colprop * 100, self.places):.{self.places}f}%)"
            )
        elif self.style == N_MEAN:
            col_value = f"{round(self.row.mean, self.places):.{self.places}f}"
        elif self.style == MEDIAN_IQR:
            col_value = (
                f"{round(self.row.q50, self.places):.{self.places}f} "
                f"({round(self.row.q25, self.places)},"
                f"{round(self.row.q75, self.places):.{self.places}f})"
            )
        elif self.style == MEDIAN_RANGE:
            col_value = (
                f"{round(self.row.q50, self.places):.{self.places}f} "
                f"({round(self.row.min, self.places):.{self.places}f}, "
                f"{round(self.row.max, self.places):.{self.places}f})"
            )
        elif self.style == MEAN_RANGE:
            col_value = (
                f"{round(self.row.mean, self.places):.{self.places}f} "
                f"({round(self.row.min, self.places):.{self.places}f}, "
                f"{round(self.row.max, self.places):.{self.places}f})"
            )
        elif self.style == MEAN_SD:
            col_value = (
                f"{round(self.row.mean, self.places):.{self.places}f} "
                f"({round(self.row.sd, self.places):.{self.places}f})"
            )
        elif self.style == MEAN_95CI:
            col_value = (
                f"{round(self.row.mean, self.places):.{self.places}f} "
                f"({round(self.row.ci95l, self.places):.{self.places}f}, "
                f"{round(self.row.ci95h, self.places):.{self.places}f})"
            )
        return col_value



from django import forms
from django.core.exceptions import ObjectDoesNotExist

from edc_registration.models import RegisteredSubject

from ...models import StockRequestItem


class StockRequestItemForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        try:
            RegisteredSubject.objects.get(
                subject_identifier=cleaned_data.get("subject_identifier"),
                consent_datetime__isnull=False,
                randomization_datetime__isnull=False,
            )
        except ObjectDoesNotExist:
            raise forms.ValidationError({"subject_identifier": "Subject does not exist"})
        return cleaned_data

    class Meta:
        model = StockRequestItem
        fields = "__all__"
        help_text = {"request_item_identifier": "(read-only)"}
        widgets = {
            "request_item_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



import uuid

from django.db import models
from django.utils import timezone

from ..constants import DELETE, INSERT, UPDATE


class ExportTrackingFieldsModelMixin(models.Model):
    """Adds these fields to the Concrete model."""

    exported = models.BooleanField(
        default=False,
        editable=False,
        help_text=(
            "system field for export tracking. considered "
            "'exported' if both sent and received."
        ),
    )

    exported_datetime = models.DateTimeField(
        null=True, editable=False, help_text="system field for export tracking."
    )

    export_change_type = models.CharField(
        max_length=1,
        choices=((INSERT, "Insert"), (UPDATE, "Update"), (DELETE, "Delete")),
        default=INSERT,
        editable=False,
        help_text="system field for export tracking.",
    )

    export_uuid = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        help_text="system field for export tracking.",
    )

    def update_export_mixin_fields(self):
        self.exported = True
        self.exported_datetime = timezone.now()
        self.save()

    class Meta:
        abstract = True



from typing import Any

from django.contrib import messages
from django.contrib.auth.models import User
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin

from ..exportables import Exportables


class ExportModelsView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = "edc_export/export_models.html"
    navbar_name = "edc_export"
    navbar_selected_item = "export"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        if self.kwargs.get("action") == "cancel":
            try:
                self.request.session.pop("selected_models")
            except KeyError:
                pass
            else:
                messages.info(self.request, "Nothing has been exported.")
        user = User.objects.get(username=self.request.user)
        kwargs.update(exportables=Exportables(request=self.request, user=user))
        return super().get_context_data(**kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .exceptions import NavbarError

if TYPE_CHECKING:
    from django.contrib.auth.models import User

    from .navbar_item import NavbarItem


@dataclass
class Navbar:
    """A class to contain a list of navbar items. See NavbarItem."""

    name: str = None
    navbar_items: list[NavbarItem] = field(default_factory=list, init=False)

    def register(self, navbar_item: NavbarItem):
        if navbar_item in self.navbar_items:
            raise NavbarError(f"Duplicate navbar item. See {self}. Got{navbar_item.name}")
        self.navbar_items.append(navbar_item)

    @property
    def names(self) -> list[str]:
        return [navbar_item.name for navbar_item in self.navbar_items]

    def get(self, name: str) -> NavbarItem | None:
        try:
            navbar_item = [nb for nb in self.navbar_items if nb.name == name][0]
        except IndexError:
            navbar_item = None
        return navbar_item

    def set_active(self, name: str) -> None:
        if name:
            for navbar_item in self.navbar_items:
                navbar_item.active = True if navbar_item.name == name else False

    def show_user_permissions(self, user: User = None) -> dict[str, dict[str, bool]]:
        """Returns the permissions required to access this Navbar
        and True if the given user has such permissions.
        """
        permissions = {}
        for navbar_item in self.navbar_items:
            has_perm = {}
            has_perm.update({navbar_item.codename: user.has_perm(navbar_item.codename)})
            permissions.update({navbar_item.name: has_perm})
        return permissions



# Generated by Django 6.0 on 2025-02-27 09:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0055_historicalreceiveitem_comment_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicallot",
            old_name="manufacture_date",
            new_name="manufactured_date",
        ),
        migrations.RenameField(
            model_name="lot",
            old_name="manufacture_date",
            new_name="manufactured_date",
        ),
        migrations.AddField(
            model_name="historicallot",
            name="comment",
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name="historicallot",
            name="country_of_origin",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="historicallot",
            name="manufactured_by",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="historicallot",
            name="processed_until_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="historicallot",
            name="reference",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="historicallot",
            name="storage_conditions",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="lot",
            name="comment",
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name="lot",
            name="country_of_origin",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="lot",
            name="manufactured_by",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="lot",
            name="processed_until_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="lot",
            name="reference",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="lot",
            name="storage_conditions",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



import sys

from django.core.management.base import BaseCommand

from ...model_exporter import ModelExporter


class Command(BaseCommand):
    help = "Export model objects history for a given app_label.modelname."

    def add_arguments(self, parser):
        parser.add_argument(
            "--models",
            dest="models",
            nargs="*",
            default=None,
            help=("run for a select list of models (label_lower syntax)"),
        )

    def handle(self, *args, **options):
        models = options.get("models")
        for model in models:
            sys.stdout.write(f" ( ) {model} ...\r")
            model_exporter = ModelExporter(model=model)
            path = model_exporter.export()
            sys.stdout.write(f" (*) {model}      \n")
            sys.stdout.write(f"     file: '{path}'\n")



MEAN_95CI = "mean_95ci"
MEAN_RANGE = "mean_range"
MEAN_SD = "mean_sd"
MEDIAN_IQR = "median_iqr"
MEDIAN_RANGE = "median_range"
N_MEAN = "n_mean"
N_ONLY = "n_only"
N_RANGE = "n_range"
N_WITH_COL_PROP = "n_with_col_prop"
N_WITH_ROW_PROP: str = "n_with_row_prop"

COUNT_COLUMN = "count"
TITLE_COLUMN = "title"

STATISTICS = [
    MEAN_95CI,
    MEAN_RANGE,
    MEAN_SD,
    MEDIAN_IQR,
    MEDIAN_RANGE,
    N_MEAN,
    N_ONLY,
    N_RANGE,
    N_WITH_COL_PROP,
    N_WITH_ROW_PROP,
]



from .constants import QUESTION_RETIRED


def get_display(choices, label) -> str | None:
    """Returns the display value of a choices tuple for label."""
    for choice in choices:
        store_value, display_value = choice
        if label == store_value:
            return display_value
    return None


def append_question_retired_choice(choices) -> tuple[tuple[str, str], ...]:
    choices = list(choices)
    choices.append((QUESTION_RETIRED, QUESTION_RETIRED))
    choices = tuple(choices)
    return choices



from django.db import models


class AliquotIdentifierModelMixin(models.Model):
    # NOTE: aliquot_identifier is allocated by the Specimen object
    aliquot_identifier = models.CharField(max_length=25, unique=True)

    parent_identifier = models.CharField(
        verbose_name="Parent aliquot Identifier", max_length=25, editable=False
    )

    identifier_prefix = models.CharField(max_length=50, editable=False)

    subject_identifier = models.CharField(max_length=50, default="", editable=False)

    requisition_identifier = models.CharField(max_length=50, default="", editable=False)

    class Meta:
        abstract = True



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.sites.management import create_default_site
from django.core.checks import register
from django.core.management.color import color_style
from django.db.models.signals import post_migrate
from multisite.apps import post_migrate_sync_alias

from edc_action_item.post_migrate_signals import update_action_types
from edc_action_item.site_action_items import site_action_items
from edc_action_item.system_checks import edc_action_item_checks
from edc_auth.post_migrate_signals import post_migrate_user_groups_and_roles
from edc_auth.site_auths import site_auths
from edc_consent.site_consents import site_consents
from edc_data_manager.post_migrate_signals import (
    populate_data_dictionary,
    update_query_rule_handlers,
)
from edc_data_manager.site_data_manager import site_data_manager
from edc_export.system_checks import edc_export_checks
from edc_facility.system_checks import holiday_country_check, holiday_path_check
from edc_form_runners.site import site_form_runners
from edc_lab.post_migrate_signals import update_panels_on_post_migrate
from edc_lab.site_labs import site_labs
from edc_list_data.post_migrate_signals import post_migrate_list_data
from edc_list_data.site_list_data import site_list_data
from edc_metadata.metadata_rules import site_metadata_rules
from edc_metadata.system_checks import check_for_metadata_rules
from edc_navbar.site_navbars import site_navbars
from edc_navbar.system_checks import edc_navbar_checks
from edc_notification.post_migrate_signals import post_migrate_update_notifications
from edc_notification.site_notifications import site_notifications
from edc_pdutils.site_values_mappings import site_values_mappings
from edc_prn.site_prn_forms import site_prn_forms
from edc_pylabels.site_label_configs import site_label_configs
from edc_randomization.site_randomizers import site_randomizers
from edc_reportable.post_migrate_signals import post_migrate_load_reference_ranges
from edc_sites.post_migrate_signals import post_migrate_update_sites
from edc_sites.site import sites as site_sites
from edc_sites.system_checks import sites_check
from edc_visit_schedule.post_migrate_signals import populate_visit_schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.system_checks import (
    check_form_collections,
    check_onschedule_exists_in_subject_schedule_history,
    check_subject_schedule_history,
    visit_schedule_check,
)

installed_apps = [x.split(".apps")[0] for x in settings.INSTALLED_APPS]

style = color_style()

__all__ = ["AppConfig"]


class AppConfig(DjangoAppConfig):
    """AppConfig class for main EDC apps.py.

    Should be the last app in INSTALLED_APPS

    The post_migrate signal(s) registered here will
    find site globals fully populated.

    For example,
    'post_migrate_user_groups_and_roles' needs site_consents
    to be fully populated before running.
    """

    name = "edc_appconfig"
    verbose_name = "Edc AppConfig"
    has_exportable_data = False
    include_in_administration_section = False

    def ready(self):
        sys.stdout.write("Loading edc_appconfig ...\n")
        self.call_autodiscovers()
        self.register_system_checks()
        self.unregister_post_migrate_signals()
        self.register_post_migrate_signals()
        sys.stdout.write("Done loading edc_appconfig.\n")

    @staticmethod
    def call_autodiscovers():
        """Call autodiscover on apps to load globals"""
        opts = dict(
            edc_consent=site_consents.autodiscover,
            edc_auth=site_auths.autodiscover,
            edc_sites=site_sites.autodiscover,
            edc_lab=site_labs.autodiscover,
            edc_list_data=site_list_data.autodiscover,
            edc_action_item=site_action_items.autodiscover,
            edc_data_manager=site_data_manager.autodiscover,
            edc_notification=site_notifications.autodiscover,
            edc_form_runners=site_form_runners.autodiscover,
            edc_metadata=site_metadata_rules.autodiscover,
            edc_visit_schedule=site_visit_schedules.autodiscover,
            edc_navbar=site_navbars.autodiscover,
            edc_pdutils=site_values_mappings.autodiscover,
            edc_prn=site_prn_forms.autodiscover,
            edc_randomization=site_randomizers.autodiscover,
            edc_pylabels=site_label_configs.autodiscover,
        )
        # site_offline_models.autodiscover()
        opts = {k: v for k, v in opts.items() if k in installed_apps}
        for app, autodiscover in opts.items():
            autodiscover()

    @staticmethod
    def register_system_checks():
        """Register system checks"""
        from edc_consent.system_checks import check_consents  # wait, app not ready

        sys.stdout.write(" * registering system checks\n")
        if "edc_visit_schedule" in installed_apps:
            sys.stdout.write("   - visit_schedule_check\n")
            register(visit_schedule_check)
            sys.stdout.write("   - check_form_collections\n")
            register(check_form_collections)
            sys.stdout.write("   - check subject schedule history\n")
            register(check_subject_schedule_history, deploy=True)
            sys.stdout.write("   - check onschedule with subject schedule history\n")
            register(check_onschedule_exists_in_subject_schedule_history)
        if "edc_action_item" in installed_apps:
            sys.stdout.write("   - edc_action_item_checks\n")
            register(edc_action_item_checks)
        if "edc_sites" in installed_apps:
            sys.stdout.write("   - sites_check\n")
            register(sites_check)
        if "edc_export" in installed_apps:
            sys.stdout.write("   - edc_export_checks\n")
            register(edc_export_checks, deploy=True)
        if "edc_facility" in installed_apps:
            sys.stdout.write("   - holiday_path_check (deploy only)\n")
            register(holiday_path_check, deploy=True)
            sys.stdout.write("   - holiday_country_check (deploy only)\n")
            register(holiday_country_check, deploy=True)
        if "edc_metadata" in installed_apps:
            sys.stdout.write("   - check_for_metadata_rules (deploy only)\n")
            register(check_for_metadata_rules)
        if "edc_consent" in installed_apps:
            sys.stdout.write("   - check_site_consents\n")
            register(check_consents)
        if "edc_navbar" in installed_apps:
            sys.stdout.write("   - edc_navbar_checks\n")
            register(edc_navbar_checks)

    def unregister_post_migrate_signals(self):
        """Unregister post-migrate signals.

        Unregister the default signal that creates "example.com"
        instead of deleting the site instance later.
        Deleting the "example.com" site instance later raises an
        OperationalError("cannot modify your_model because it is a
        view"). The exception is refering to any unmanaged model
        (your_model) based on SQL VIEWS that has a FK to model Sites.

        See also: edc_qareports.models.dbviews
        """
        if not getattr(settings, "EDC_SITES_CREATE_DEFAULT", True):
            sys.stdout.write(
                " * unregistering django post-migrate signal `create_default_site` ...\n"
            )
            post_migrate.disconnect(
                create_default_site, sender=django_apps.get_app_config("sites")
            )

    def register_post_migrate_signals(self):
        """Register post-migrate signals."""
        sys.stdout.write(" * registering post-migrate signals ...\n")
        if "edc_visit_schedule" in installed_apps:
            sys.stdout.write("   - post_migrate.populate_visit_schedule\n")
            post_migrate.connect(
                populate_visit_schedule,
                sender=self,
                dispatch_uid="edc_visit_schedule.populate_visit_schedule",
            )
        if "edc_sites" in installed_apps:
            sys.stdout.write("   - post_migrate.post_migrate_update_sites\n")
            post_migrate.connect(
                post_migrate_update_sites,
                sender=self,
                dispatch_uid="edc_sites.post_migrate_update_sites",
            )
        if "multisite" in installed_apps:
            sys.stdout.write("   - post_migrate.multisite.post_migrate_sync_alias\n")
            post_migrate.connect(
                post_migrate_sync_alias,
                sender=self,
                dispatch_uid="multisite.post_migrate_sync_alias",
            )
        if "edc_lab" in installed_apps:
            sys.stdout.write("   - post_migrate.update_panels_on_post_migrate\n")
            post_migrate.connect(
                update_panels_on_post_migrate,
                sender=self,
                dispatch_uid="edc_lab.update_panels_on_post_migrate",
            )
        if "edc_list_data" in installed_apps:
            sys.stdout.write("   - post_migrate.post_migrate_list_data\n")
            post_migrate.connect(
                post_migrate_list_data,
                sender=self,
                dispatch_uid="edc_list_data.post_migrate_list_data",
            )
        if "edc_action_item" in installed_apps:
            sys.stdout.write("   - post_migrate.update_action_types\n")
            post_migrate.connect(
                update_action_types,
                sender=self,
                dispatch_uid="edc_action_item.update_action_types",
            )
        if "edc_auth" in installed_apps:
            sys.stdout.write("   - post_migrate.post_migrate_user_groups_and_roles\n")
            post_migrate.connect(
                post_migrate_user_groups_and_roles,
                sender=self,
                dispatch_uid="edc_auth.post_migrate_user_groups_and_roles",
            )
        if "edc_data_manager" in installed_apps:
            sys.stdout.write("   - post_migrate.update_query_rule_handlers\n")
            post_migrate.connect(
                update_query_rule_handlers,
                sender=self,
                dispatch_uid="edc_data_manager.update_query_rule_handlers",
            )
            sys.stdout.write("   - post_migrate.populate_data_dictionary\n")
            post_migrate.connect(
                populate_data_dictionary,
                sender=self,
                dispatch_uid="edc_data_manager.populate_data_dictionary",
            )
        if "edc_notification" in installed_apps:
            sys.stdout.write("   - post_migrate.post_migrate_update_notifications\n")
            post_migrate.connect(
                post_migrate_update_notifications,
                sender=self,
                dispatch_uid="edc_notification.post_migrate_update_notifications",
            )
        if "edc_reportable" in installed_apps:
            sys.stdout.write("   - post_migrate.post_migrate_load_reference_ranges\n")
            post_migrate.connect(
                post_migrate_load_reference_ranges,
                sender=self,
                dispatch_uid="edc_reportable.post_migrate_load_reference_ranges",
            )



from datetime import date, datetime, timedelta

from django.core.exceptions import ValidationError
from django.utils import timezone


def datetime_not_future(utc_datetime: datetime) -> None:
    time_error = timedelta(minutes=10)
    if utc_datetime > timezone.now() + time_error:
        raise ValidationError("Cannot be a future date/time")


def date_not_future(value: date) -> None:
    if value > timezone.now().date():
        raise ValidationError("Cannot be a future date")


def date_is_past(value: date) -> None:
    if value > timezone.now().date():
        raise ValidationError("Expected a past date")


def date_is_not_now(value: date) -> None:
    if value == timezone.now().date():
        raise ValidationError("Cannot be today")


def datetime_is_future(utc_datetime: datetime) -> None:
    time_error = timedelta(minutes=10)
    if utc_datetime < timezone.now() + time_error:
        raise ValidationError("Expected a future date/time")


def date_is_future(value: date) -> None:
    if value < timezone.now().date():
        raise ValidationError("Expected a future date")



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.db import models
from django.db.models import QuerySet

from .constants import CRF, KEYED, REQUISITION

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_crf.model_mixins import CrfModelMixin as Base

    from .model_mixins.creates import CreatesMetadataModelMixin

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        appointment: Appointment

    from edc_metadata.models import CrfMetadata, RequisitionMetadata

    from .model_mixins.updates import (
        UpdatesCrfMetadataModelMixin,
        UpdatesRequisitionMetadataModelMixin,
    )

    class CrfModel(UpdatesCrfMetadataModelMixin, Base):
        related_visit = models.ForeignKey(RelatedVisitModel, on_delete=models.PROTECT)

    class RequisitionModel(UpdatesRequisitionMetadataModelMixin, Base):
        related_visit = models.ForeignKey(RelatedVisitModel, on_delete=models.PROTECT)

    class ScheduledLikeModel:
        subject_identifier: str
        visit_code: str
        visit_code_sequence: int
        visit_schedule_name: str
        schedule_name: str


class HasKeyedMetadata(Exception):
    pass


def get_crf_metadata_model_cls() -> type[CrfMetadata]:
    return django_apps.get_model("edc_metadata.crfmetadata")


def get_requisition_metadata_model_cls() -> type[RequisitionMetadata]:
    return django_apps.get_model("edc_metadata.requisitionmetadata")


def get_metadata_model_cls(
    metadata_category: str,
) -> type[CrfMetadata] | type[RequisitionMetadata]:
    if metadata_category == CRF:
        model_cls = get_crf_metadata_model_cls()
    elif metadata_category == REQUISITION:
        model_cls = get_requisition_metadata_model_cls()
    else:
        raise ValueError(f"Invalid metadata category. Got {metadata_category}.")
    return model_cls


def verify_model_cls_registered_with_admin():
    return getattr(settings, "EDC_METADATA_VERIFY_MODELS_REGISTERED_WITH_ADMIN", False)


def refresh_metadata_for_timepoint(
    instance: CrfModel | RequisitionModel | Appointment | RelatedVisitModel,
    allow_create: bool | None = None,
):
    """Refresh (or creates) metadata for the given timepoint.

    See also `metadata_create_on_post_save` and `CreatesMetadataModelMixin`.
    """
    if instance:
        try:
            related_visit = instance.related_visit
        except AttributeError:
            related_visit = instance
        if allow_create:
            related_visit.metadata_create()
        if django_apps.get_app_config("edc_metadata").metadata_rules_enabled:
            related_visit.run_metadata_rules(allow_create=allow_create)


def get_crf_metadata(
    instance: ScheduledLikeModel | Appointment,
) -> QuerySet[CrfMetadata]:
    """Returns a queryset of crf metedata."""
    opts = dict(
        subject_identifier=instance.subject_identifier,
        visit_schedule_name=instance.visit_schedule_name,
        schedule_name=instance.schedule_name,
        visit_code=instance.visit_code,
        visit_code_sequence=instance.visit_code_sequence,
    )
    return get_crf_metadata_model_cls().objects.filter(**opts)


def get_requisition_metadata(
    instance: ScheduledLikeModel | Appointment,
) -> QuerySet[RequisitionMetadata]:
    """Returns a queryset of requisition metadata"""
    opts = dict(
        subject_identifier=instance.subject_identifier,
        visit_schedule_name=instance.visit_schedule_name,
        schedule_name=instance.schedule_name,
        visit_code=instance.visit_code,
        visit_code_sequence=instance.visit_code_sequence,
    )
    return get_requisition_metadata_model_cls().objects.filter(**opts)


def has_keyed_metadata(appointment, raise_on_true=None) -> bool:
    """Return True if data has been submitted for this timepoint."""
    exists = any(
        [
            get_crf_metadata(appointment).filter(entry_status=KEYED).exists(),
            get_requisition_metadata(appointment).filter(entry_status=KEYED).exists(),
        ]
    )
    if exists and raise_on_true:
        raise HasKeyedMetadata(
            "Metadata data exists for this timepoint. Got "
            f"{appointment.visit_code}.{appointment.visit_code_sequence}."
        )
    return exists



from .ae_followup_admin_mixin import AeFollowupModelAdminMixin
from .ae_initial_admin_mixin import (
    AeInitialModelAdminMixin,
    default_radio_fields,
    fieldset_part_four,
    fieldset_part_one,
    fieldset_part_three,
)
from .ae_susar_admin_mixin import AeSusarModelAdminMixin
from .ae_tmg_admin_mixin import AeTmgModelAdminMixin
from .death_report_admin_mixin import DeathReportModelAdminMixin
from .death_report_tmg_admin_mixin import DeathReportTmgModelAdminMixin
from .hospitalization_admin_mixin import HospitalizationModelAdminMixin
from .modeladmin_mixins import AdverseEventModelAdminMixin, NonAeInitialModelAdminMixin



# Generated by Django 2.1.2 on 2018-11-08 01:53

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0015_auto_20181010_2125")]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        )
    ]



from edc_action_item.modelform_mixins import ActionItemModelFormMixin
from edc_form_validators import FormValidatorMixin
from edc_model_form.mixins import BaseModelFormMixin
from edc_offstudy.modelform_mixins import OffstudyNonCrfModelFormMixin
from edc_sites.forms import SiteModelFormMixin


class AeModelFormMixin(
    SiteModelFormMixin,
    OffstudyNonCrfModelFormMixin,
    ActionItemModelFormMixin,
    FormValidatorMixin,
    BaseModelFormMixin,
):
    pass



from __future__ import annotations

import pandas as pd
from django.apps import apps as django_apps
from django.contrib.sites.models import Site
from django.db import models
from django_pandas.io import read_frame

from edc_model_to_dataframe.read_frame_edc import read_frame_edc
from edc_registration.models import RegisteredSubject

from ..constants import SYSTEM_COLUMNS
from ..utils import (
    convert_dates_from_model,
    convert_numbers_to_nullable_dtype,
    convert_timedelta_from_model,
)
from .get_subject_visit import get_subject_visit


def get_crf(
    model: str | None = None,
    subject_visit_model: str | None = None,
    drop_columns: list[str] | None = None,
    subject_identifiers: list[str] | None = None,
    normalize: bool | None = None,
    localize: bool | None = None,
    read_verbose: bool | None = None,
    drop_sys_columns: bool | None = None,
    drop_action_item_columns: bool | None = None,
) -> pd.DataFrame:
    """Return a dataframe of CRF model.

    Merge with subject visit model of model lanel_lower
    id provided.

    Rename columns site to site_id, subject_visit to subject_visit_id
    """

    normalize = True if normalize is None else normalize
    localize = True if localize is None else localize
    read_verbose = True if read_verbose is None else read_verbose
    model_cls = django_apps.get_model(model)
    if subject_identifiers:
        qs = model_cls.objects.filter(
            subject_visit__subject_identifier__in=subject_identifiers
        )
    else:
        qs = model_cls.objects.all()

    df = read_frame(qs, verbose=read_verbose)
    if read_verbose:
        sites = {obj.domain: obj.id for obj in Site.objects.all()}
        df["site"] = df["site"].map(sites)
    df = df.rename(columns={"site": "site_id", "subject_visit": "subject_visit_id"})

    if "subject_visit_id" not in df.columns:
        raise ValueError("This is not a CRF. Requires col subject_visit_id.")

    df = df.reset_index(drop=True)
    df_subject_visit = get_subject_visit(
        subject_visit_model, subject_identifiers=subject_identifiers
    )
    df = pd.merge(
        df_subject_visit,
        df,
        on="subject_visit_id",
        how="right",
        suffixes=("", "_subject_visit"),
    )
    df = df.drop(columns=[col for col in df.columns if col.endswith("_subject_visit")])
    df = df.reset_index(drop=True)

    df["subject_visit_id_original"] = df["subject_visit_id"]
    df["subject_visit_id"] = df["subject_visit_id"].astype(str)
    for field in model_cls._meta.get_fields():
        if isinstance(field, models.ManyToManyField):
            df_m2m = (
                read_frame_edc(
                    qs,
                    read_frame_verbose=False,
                    drop_sys_columns=drop_sys_columns,
                    drop_action_item_columns=drop_action_item_columns,
                )
                .sort_values(["subject_identifier", "visit_code"])
                .reset_index(drop=True)
            )
            df_m2m["subject_visit_id"] = df_m2m["subject_visit_id"].astype(str)
            df = df.merge(
                df_m2m[["subject_visit_id", field.name]],
                on="subject_visit_id",
                how="left",
            )
    df = df.drop(columns=["subject_visit_id"])
    df = df.rename(columns={"subject_visit_id_original": "subject_visit_id"})

    # move system columns to end
    df = df[[col for col in df.columns if col not in SYSTEM_COLUMNS] + SYSTEM_COLUMNS]
    if drop_columns:
        df = df.drop(columns=drop_columns)

    # add demographics
    df_rs = read_frame(
        RegisteredSubject.objects.values(
            "subject_identifier", "gender", "dob", "ethnicity"
        ).all(),
        verbose=read_verbose,
    )
    df = df.merge(df_rs, on="subject_identifier", how="left")
    first_cols = [
        "subject_identifier",
        "gender",
        "dob",
        "ethnicity",
        "appointment_id",
        "subject_visit_id",
    ]
    df = df[
        [
            *first_cols,
            *[col for col in df.columns if col not in first_cols],
        ]
    ]

    # convert values to ...
    # df = convert_numerics_from_model(df, model_cls)
    df = convert_numbers_to_nullable_dtype(df)
    df = convert_dates_from_model(df, model_cls, normalize=normalize, localize=localize)
    df = convert_timedelta_from_model(df, model_cls)
    df = df.replace("", pd.NA).fillna(pd.NA).reset_index(drop=True)
    return df



from .dashboard_model_button import DashboardModelButton
from .history_button import HistoryButton
from .model_button import ADD, CHANGE, VIEW, ModelButton
from .next_querystring import NextQuerystring
from .perms import Perms
from .prn_button import PrnButton
from .query_button import QueryButton
from .render_history_and_query_buttons import render_history_and_query_buttons

__all__ = [
    "ADD",
    "CHANGE",
    "VIEW",
    "DashboardModelButton",
    "HistoryButton",
    "ModelButton",
    "NextQuerystring",
    "Perms",
    "PrnButton",
    "QueryButton",
    "render_history_and_query_buttons",
]



from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register
from django.core.management.color import color_style

from .system_checks import edc_notification_check

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_notification"
    verbose_name = "Edc Notification"
    include_in_administration_section = True

    def ready(self):
        register(edc_notification_check)



from django.conf import settings

from edc_constants.constants import LIVE, TEST
from edc_protocol.research_protocol_config import ResearchProtocolConfig


def admin_theme(request) -> dict:
    dct = {
        "LIVE_SYSTEM": getattr(settings, "LIVE_SYSTEM", False),
        "DEBUG": getattr(settings, "DEBUG", False),
        "LIVE": LIVE,
        "TEST": TEST,
        "project_name": ResearchProtocolConfig().project_name,
    }
    if theme := getattr(settings, "EDC_MODEL_ADMIN_CSS_THEME", None):
        dct.update(
            {
                "edc_model_admin_css_theme_path": (
                    [
                        "edc_model_admin/admin/css/edc_model_admin.css",
                        f"edc_model_admin/admin/css/themes/{theme.lower()}.css",
                    ]
                )
            }
        )
    return dct



from .fields import (
    DiastolicPressureField,
    HeartRateField,
    HeightField,
    RespiratoryRateField,
    SystolicPressureField,
    TemperatureField,
    WaistCircumferenceField,
    WeightField,
)



import copy
import inspect
from typing import Any

from .rule import Rule
from .rule_group_meta_options import RuleGroupMetaOptions


class RuleGroupError(Exception):
    pass


class RuleGroupMetaclass(type):
    """Rule group metaclass."""

    rule_group_meta = RuleGroupMetaOptions

    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> Any:
        try:
            abstract = attrs.get("Meta", False).abstract
        except AttributeError:
            abstract = False
        parents = [b for b in bases if isinstance(b, RuleGroupMetaclass)]
        if not parents or abstract:
            # If this isn't a subclass, don't do anything special.
            return super().__new__(mcs, name, bases, attrs)

        # get rules from abstract parents
        for parent in parents:
            try:
                if parent.Meta.abstract:
                    for rule in [
                        member
                        for member in inspect.getmembers(parent)
                        if isinstance(member[1], Rule)
                    ]:
                        parent_rule = copy.deepcopy(rule)
                        attrs.update({parent_rule[0]: parent_rule[1]})
            except AttributeError:
                pass

        # prepare "meta" options from class Meta:
        meta = mcs.rule_group_meta(name, attrs)
        # add the rules tuple to the meta options
        rules = mcs.__get_rules(name, attrs, meta)
        meta.options.update(rules=rules)
        # ... django style _meta
        attrs.update({"_meta": meta})

        attrs.update({"name": f"{meta.app_label}.{name.lower()}"})
        return super().__new__(mcs, name, bases, attrs)

    @classmethod
    def __get_rules(mcs, name: str, attrs: dict, meta: Any) -> tuple:
        """Returns a list of rules after updating each rule's attrs
        with values from Meta.

        Note: Attrs from Meta will overwrite those on rule.
        """
        rules = []
        for key, value in attrs.items():
            if not key.startswith("_"):
                if isinstance(value, Rule):
                    rule = value
                    rule.name = key
                    rule.group = name
                    if isinstance(rule.predicate, (str,)):
                        predicates = getattr(meta, "predicates", None)
                        if not predicates:
                            raise RuleGroupError(
                                "RuleGroup Meta attr `predicates` may not be `None` if a "
                                "rule.predicate in the RuleGroup is a string. "
                                f"See {attrs.get('__qualname__')}."
                            )
                        rule.predicate = getattr(meta.predicates, rule.predicate)
                    for k, v in meta.options.items():
                        setattr(rule, k, v)
                    rule.target_models = mcs.__get_target_models(rule, meta)
                    rules.append(rule)
        return tuple(rules)

    @classmethod
    def __get_target_models(mcs, rule: Any, meta: Any) -> Any:
        """Returns target models as a list of label_lowers.

        Target models are the models whose metadata is acted upon.

        If `model_name` instead of `label_lower`, assumes `app_label`
        from meta.app_label.
        """
        target_models = []
        for target_model in rule.target_models:
            if len(target_model.split(".")) != 2:
                target_model = f"{meta.app_label}.{target_model}"
            target_models.append(target_model)
        return target_models



from .drawing_callable_example import draw_callable_example
from .site_label_configs import site_label_configs


class Container:
    qty = 128


class StockRequest:
    container = Container()


class Site:
    id = 10
    name = "Amana"


class RequestItem:
    subject_identifier = "999-99-99999-9"
    gender = "FEMALE"
    site = Site
    stock_request = StockRequest()
    code = "A9B8C7"
    sid = 99999


site_label_configs.register("test_label_config", draw_callable_example, RequestItem)



# Generated by Django 5.2.1 on 2025-06-04 20:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_reportable", "0004_alter_referencerangecollection_grade3_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalnormaldata",
            name="auto_created",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="normaldata",
            name="auto_created",
            field=models.BooleanField(default=False),
        ),
    ]



import copy
from collections import OrderedDict


class FieldsetError(Exception):
    pass


class Fieldsets:
    """A class to use with model admin fieldsets."""

    def __init__(self, fieldsets=None, **kwargs):
        self.fieldsets_asdict = OrderedDict(copy.deepcopy(fieldsets))

    @property
    def fieldsets(self):
        return tuple([(k, v) for k, v in self.fieldsets_asdict.items()])

    def move_to_end(self, sections):
        if sections:
            for section in sections:
                self.fieldsets_asdict.move_to_end(section)

    def add_fieldset(self, section=None, fields=None, fieldset=None):
        """Adds a fieldset to the given section/fields or fieldset.

        * fieldset: a Fieldset class instance.
        """
        if fieldset:
            section = fieldset.fieldset[0]
            fields = fieldset.fieldset[1]["fields"]
        self.fieldsets_asdict.update({section: {"fields": fields}})

    def add_fieldsets(self, fieldsets=None):
        """Adds a list of fieldsets."""
        try:
            fieldsets[0]
        except TypeError:
            fieldsets = [fieldsets]
        for fieldset in fieldsets:
            section = fieldset.fieldset[0]
            fields = fieldset.fieldset[1]["fields"]
            self.fieldsets_asdict.update({section: {"fields": fields}})

    def insert_fields(self, *insert_fields, insert_after=None, section=None):
        """Inserts fields after insert_after in the given section."""
        if insert_fields and insert_fields != (None,):
            fields = self._copy_section_fields(section)
            position = self._get_field_position(fields, insert_after)
            for index, field in enumerate(insert_fields):
                fields.insert(index + position, field)
            self.fieldsets_asdict[section]["fields"] = tuple(fields)

    def remove_fields(self, *remove_fields, section=None):
        """Removes fields from the given section."""
        if remove_fields and remove_fields != (None,):
            fields = self._copy_section_fields(section)
            fields = [f for f in fields if f not in remove_fields]
            self.fieldsets_asdict[section]["fields"] = tuple(fields)

    def _copy_section_fields(self, section):
        """Returns fields as a list which is a copy of the fields
        tuple in a section or raises if section does not exist.
        """
        try:
            fields = copy.copy(self.fieldsets_asdict[section]["fields"])
        except KeyError:
            raise FieldsetError(f"Invalid fieldset section. Got {section}")
        return list(fields)

    @staticmethod
    def _get_field_position(fields, insert_after):
        try:
            position = fields.index(insert_after) + 1
        except ValueError:
            raise FieldsetError(
                f"Field does not exist in section {fields}. Got {insert_after}"
            )
        return position



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from django.core.management.color import color_style
from django.urls import NoReverseMatch
from django.urls.base import reverse

from edc_dashboard.url_names import InvalidDashboardUrlName, url_names

if TYPE_CHECKING:
    from django.contrib.auth.models import User

style = color_style()


@dataclass
class NavbarItem:
    """A class that represents a single item on a navbar."""

    name: str = field(default=None, compare=True)
    title: str = field(default=None)
    label: str | None = field(default=None)
    codename: str = field(default=None)
    url_name: str = field(default=None)
    no_url_namespace: bool = field(default=None)
    fa_icon: str | None = field(default=None)
    disabled: str = field(default="disabled")

    active: bool = field(default=None)

    template_name: str = field(
        default="edc_navbar/navbar_item.html",
        repr=False,
    )

    def __post_init__(self):
        self.title = self.title or self.label or self.name.title()  # the anchor title

    def get_url(self, raise_exception: bool | None = None) -> str | None:
        try:
            url = reverse(self.real_url_name)
        except NoReverseMatch:
            url = None
            if raise_exception:
                raise
        return url

    @property
    def real_url_name(self) -> str:
        try:
            url_name = url_names.get(self.url_name)
        except InvalidDashboardUrlName:
            url_name = self.url_name.split(":")[1] if self.no_url_namespace else self.url_name
        return url_name

    def set_disabled(self, user: User | None = None):
        if user and user.has_perm(self.codename):
            self.disabled = ""



glucose_fieldset_tuple: tuple[str, dict[str, tuple]] = (
    "Blood Glucose Measurement",
    {
        "fields": (
            "glucose_performed",
            "glucose_fasting",
            "glucose_fasting_duration_str",
            "glucose_date",
            "glucose_value",
            "glucose_quantifier",
            "glucose_units",
        ),
    },
)



# Generated by Django 5.1.2 on 2024-11-19 04:10

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0044_location_contact_email_location_contact_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicallot",
            name="manufacture_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="historicalreceive",
            name="invoice_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="historicalreceive",
            name="invoice_number",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="historicalreceive",
            name="supplier",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.supplier",
                verbose_name="Supplier",
            ),
        ),
        migrations.AddField(
            model_name="lot",
            name="manufacture_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="receive",
            name="invoice_date",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="receive",
            name="invoice_number",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name="receive",
            name="supplier",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.supplier",
                verbose_name="Supplier",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="lot",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.lot",
                verbose_name="Batch",
            ),
        ),
        migrations.AlterField(
            model_name="location",
            name="contact_email",
            field=models.EmailField(blank=True, max_length=150, null=True),
        ),
        migrations.AlterField(
            model_name="location",
            name="contact_tel",
            field=models.CharField(
                blank=True,
                max_length=150,
                null=True,
                validators=[django.core.validators.RegexValidator("[0-9]{1,15}")],
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="lot",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.lot",
                verbose_name="Batch",
            ),
        ),
    ]



# Generated by Django 4.2.1 on 2023-07-05 02:16

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0016_auto_20220929_1742"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="box",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Box",
                "verbose_name_plural": "Boxes",
            },
        ),
        migrations.AlterModelOptions(
            name="dispensinghistory",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Dispensing history",
                "verbose_name_plural": "Dispensing history",
            },
        ),
        migrations.AlterModelOptions(
            name="dosageguideline",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Dosage Guideline",
                "verbose_name_plural": "Dosage Guidelines",
            },
        ),
        migrations.AlterModelOptions(
            name="formulation",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Formulation",
                "verbose_name_plural": "Formulations",
            },
        ),
        migrations.AlterModelOptions(
            name="genericcontainer",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Item",
                "verbose_name_plural": "Items",
            },
        ),
        migrations.AlterModelOptions(
            name="labels",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Label",
                "verbose_name_plural": "Labels",
            },
        ),
        migrations.AlterModelOptions(
            name="location",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Location",
                "verbose_name_plural": "Locations",
            },
        ),
        migrations.AlterModelOptions(
            name="medication",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication",
                "verbose_name_plural": "Medications",
            },
        ),
        migrations.AlterModelOptions(
            name="medicationlot",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication lot",
                "verbose_name_plural": "Medication lots",
            },
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication order",
                "verbose_name_plural": "Medication orders",
            },
        ),
        migrations.AlterModelOptions(
            name="pillbottle",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Pill Bottle",
                "verbose_name_plural": "Pill Bottles",
            },
        ),
        migrations.AlterModelOptions(
            name="product",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication product",
                "verbose_name_plural": "Medication products",
            },
        ),
        migrations.AlterModelOptions(
            name="returnhistory",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Return history",
                "verbose_name_plural": "Return history",
            },
        ),
        migrations.AlterModelOptions(
            name="room",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Room",
                "verbose_name_plural": "Rooms",
            },
        ),
        migrations.AlterModelOptions(
            name="rx",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Prescription",
                "verbose_name_plural": "Prescriptions",
            },
        ),
        migrations.AlterModelOptions(
            name="rxrefill",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "RX refill",
                "verbose_name_plural": "RX refills",
            },
        ),
        migrations.AlterModelOptions(
            name="shelf",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Shelf",
                "verbose_name_plural": "Shelves",
            },
        ),
        migrations.AlterModelOptions(
            name="stock",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication stock",
                "verbose_name_plural": "Medication stock",
            },
        ),
        migrations.AlterModelOptions(
            name="stockcreatelabels",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication stock: Create labels",
                "verbose_name_plural": "Medication stock: Create labels",
            },
        ),
        migrations.AlterModelOptions(
            name="stockreceiving",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Medication stock: Receiving",
                "verbose_name_plural": "Medication stock: Receiving",
            },
        ),
        migrations.AlterModelOptions(
            name="subjectpillbottle",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Subject Pill Bottle",
                "verbose_name_plural": "Subject Pill Bottles",
            },
        ),
    ]



from edc_auth.site_auths import site_auths

from .auth_objects import (
    EDC_DX_REVIEW,
    EDC_DX_REVIEW_SUPER,
    EDC_DX_REVIEW_VIEW,
    codenames,
)

site_auths.add_group(*codenames, name=EDC_DX_REVIEW_VIEW, view_only=True)
site_auths.add_group(*codenames, name=EDC_DX_REVIEW, no_delete=True)
site_auths.add_group(*codenames, name=EDC_DX_REVIEW_SUPER)



from .data_request_admin import DataRequestAdmin
from .data_request_history_admin import DataRequestHistoryAdmin

# from .export_receipt_admin import ExportReceiptAdmin
# from .file_history_admin import FileHistoryAdmin
# from .object_history_admin import ObjectHistoryAdmin
# from .plan_admin import PlanAdmin
# from .upload_export_receipt_file_admin import UploadExportReceiptFileAdmin



from django.db import models

from .model_exporter import ObjectHistoryCreator


class ExportHistoryManager(models.Manager):
    obj_creator = ObjectHistoryCreator()

    def serialize_to_exported_object_history(
        self, instance, change_type, using, force_export=False
    ):
        """Serialize this instance to the export transaction model if ready.

        Be sure to inspect model property ready_to_export_transaction.
        ready_to_export_transaction can return True or False. If False,
        the tx will not be exported.

        if model method :func:`ready_to_export_transaction` has not been defined,
        export will proceed.

        .. note:: If change_type == 'D', entire tx is still sent."""
        try:
            ready_to_export_transaction = force_export or instance.ready_to_export_transaction
        except AttributeError as attribute_error:
            if str(attribute_error).endswith("has no attribute 'ready_to_export_transaction'"):
                ready_to_export_transaction = True
            else:
                raise
        if ready_to_export_transaction:
            self.obj_creator.create(model_obj=instance, change_type=change_type, using=using)



from django.db import models


class NotificationPlanMixin(models.Model):
    name = models.CharField(max_length=50, unique=True)

    friendly_name = models.CharField(max_length=50)

    subject_format = models.TextField()

    body_format = models.TextField()

    recipient_list = models.TextField()

    cc_list = models.TextField()

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import date, datetime

from django.conf import settings

from edc_constants.constants import EQ, GT, GTE, LT, LTE
from edc_model import estimated_date_from_ago
from edc_utils import convert_php_dateformat
from edc_utils.date import to_local

from .base_form_validator import INVALID_ERROR, BaseFormValidator


class DateValidatorError(Exception):
    pass


class DateValidator(BaseFormValidator):
    def _date_is(
        self,
        op,
        field: str | None = None,
        reference_field: str = None,
        field_value: datetime | date | None = None,
        reference_value: datetime | date | None = None,
        msg: str | None = None,
    ):
        operators = [LT, GT, EQ, LTE, GTE]
        if op not in operators:
            raise TypeError(f"Invalid operator. Expected on of {operators}.")
        if field and field_value:
            raise TypeError("Expected field name or field value but not both.")
        if reference_field and reference_value:
            raise TypeError(
                "Expected reference field name or reference field value but not both."
            )
        reference_value = reference_value or self._get_as_date(
            reference_field or "report_datetime"
        )
        field_value = field_value or self._get_as_date(field)
        if field_value and reference_value:
            if not self._compare_date_to_reference_value(op, field_value, reference_value):
                if field:
                    self.raise_validation_error({field: msg}, INVALID_ERROR)
                else:
                    self.raise_validation_error(msg, INVALID_ERROR)

    @staticmethod
    def _compare_date_to_reference_value(
        op: str,
        field_value: date,
        reference_value: date,
    ) -> bool:
        if op == LT:
            value = field_value < reference_value
        elif op == LTE:
            value = field_value <= reference_value
        elif op == GT:
            value = field_value > reference_value
        elif op == GTE:
            value = field_value >= reference_value
        elif op == EQ:
            value = field_value == reference_value
        else:
            raise DateValidatorError(f"Unknown operator. Got {op}.")
        return value

    def _get_as_date(self, key_or_value: str | date | datetime) -> date | None:
        """Returns a date or None using either a key or value.

        * If key is given, gets value from cleaned data.
        * Ensures a datetime is localized before converting to date.
        """
        try:
            value = self.cleaned_data.get(key_or_value)
        except KeyError:
            value = key_or_value
        if isinstance(value, (str,)):
            value = estimated_date_from_ago(
                cleaned_data=self.cleaned_data, ago_field=key_or_value
            )
        if value:
            try:
                value.date()
            except AttributeError:
                pass
            else:
                value = to_local(value).date()
        return value

    def date_is_after_or_raise(
        self,
        field=None,
        reference_field=None,
        field_value: datetime | date | None = None,
        reference_value: datetime | date | None = None,
        inclusive: bool = None,
        msg=None,
        extra_msg=None,
    ):
        """Raises if date/datetime field is not future to reference_field."""

        msg = msg or self.get_msg(
            "after",
            reference_value=reference_value,
            reference_field=reference_field,
            extra_msg=extra_msg,
            inclusive=inclusive,
        )
        self._date_is(
            GTE if inclusive else GT,
            field=field,
            reference_field=reference_field,
            field_value=field_value,
            reference_value=reference_value,
            msg=msg,
        )

    def date_is_before_or_raise(
        self,
        field=None,
        reference_field=None,
        field_value: datetime | date | None = None,
        reference_value: datetime | date | None = None,
        inclusive: bool = None,
        msg=None,
        extra_msg=None,
    ):
        """Raises if date/datetime field is not before the reference_field."""
        msg = msg or self.get_msg(
            "before",
            reference_value=reference_value,
            reference_field=reference_field,
            extra_msg=extra_msg,
            inclusive=inclusive,
        )
        self._date_is(
            LTE if inclusive else LT,
            field=field,
            reference_field=reference_field,
            field_value=field_value,
            reference_value=reference_value,
            msg=msg,
        )

    @staticmethod
    def get_msg(
        word: str,
        reference_value: date | datetime | None = None,
        reference_field: str | None = None,
        extra_msg: str | None = None,
        inclusive: bool | None = None,
    ) -> str:
        formatted_reference = (
            reference_value.strftime(convert_php_dateformat(settings.DATE_FORMAT))
            if reference_value
            else None
        )
        if inclusive:
            phrase = f"Expected a date on or {word}"
        else:
            phrase = f"Expected a date {word}"
        return (
            f"Invalid. {phrase} `{formatted_reference or reference_field}`. "
            f"{extra_msg or ''}".strip()
        )

    def date_is_equal_or_raise(
        self,
        field=None,
        reference_field=None,
        field_value: datetime | date | None = None,
        reference_value: datetime | date | None = None,
        msg=None,
        extra_msg=None,
    ):
        """Raises if date/datetime field is not equal the reference_field."""
        msg = msg or f"Invalid. Expected dates to match. {extra_msg or ''}".strip()
        self._date_is(
            EQ,
            field=field,
            reference_field=reference_field,
            field_value=field_value,
            reference_value=reference_value,
            msg=msg,
        )

    def date_before_report_datetime_or_raise(
        self,
        field=None,
        report_datetime_field=None,
        inclusive: bool = None,
    ):
        """Convenience method if comparing with report_datetime."""
        msg = None
        report_datetime_field = report_datetime_field or "report_datetime"
        if self.cleaned_data.get(field) and self.cleaned_data.get(report_datetime_field):
            dte = self.cleaned_data.get(report_datetime_field).strftime(
                convert_php_dateformat(settings.DATETIME_FORMAT)
            )
            phrase = "Must be on or" if inclusive else "Must be"
            msg = f"Invalid. {phrase} before report date/time. Got {dte}"
        return self.date_is_before_or_raise(
            field=field,
            reference_field=report_datetime_field,
            inclusive=inclusive,
            msg=msg,
        )

    def date_after_report_datetime_or_raise(
        self,
        field=None,
        report_datetime_field=None,
        inclusive: bool = None,
    ):
        """Convenience method if comparing with report_datetime."""
        msg = None
        report_datetime_field = report_datetime_field or "report_datetime"
        if self.cleaned_data.get(field) and self.cleaned_data.get(report_datetime_field):
            dte = self.cleaned_data.get(report_datetime_field).strftime(
                convert_php_dateformat(settings.DATETIME_FORMAT)
            )
            phrase = "Must be on or" if inclusive else "Must be"
            msg = f"Invalid. {phrase} after report date/time. Got {dte}"
        return self.date_is_after_or_raise(
            field=field,
            reference_field=report_datetime_field,
            inclusive=inclusive,
            msg=msg,
        )



from __future__ import annotations

import pandas as pd

from .stock_for_subjects import stock_for_subjects_df


def no_stock_for_subjects_df() -> pd.DataFrame:
    return stock_for_subjects_df().query("codes.isna()").reset_index(drop=True)
